"""
This script is a utility for interactively annotating data categories in the dataset manifest
"""
from typing import List, Union

import click
from fideslang import FidesModel, manifests
from fideslang.manifests import ingest_manifests
from fideslang.models import Dataset, DatasetCollection, DatasetField, FidesKey
from fideslang.validation import FidesValidationError

from fides.core import api_helpers
from fides.core.config import FidesConfig
from fides.core.utils import echo_green, get_all_level_fields


class AnnotationAbortError(Exception):
    """
    Custom exception to handle mid annotation abort
    """


def validate_data_categories(
    categories: List[str], valid_categories: List[str]
) -> None:
    """
    Check to see if each category in a list is valid
    Args:
        categories: list of category strings to check
        valid_categories: list of valid categories

    Returns:

    """
    for category in categories:
        FidesKey.validate(category)
        if category not in valid_categories:
            raise ValueError


def get_data_categories_annotation(
    dataset_member: Union[Dataset, DatasetCollection, DatasetField],
    valid_categories: List[str],
    validate: bool = True,
) -> List[str]:
    """
    Request the user's input to supply a list of data categories

    Args:
        dataset_member: A member of a dataset [dataset, collection, field]
        valid_categories: a list of the valid data categories
        validate: flag to check user inputs for formatting and data category presence

    Returns:
        List of the user's input
    """
    msg = f"""Enter comma separated data categories for [{dataset_member.name}] [s: skip, q: quit]"""
    user_response = [input.strip() for input in click.prompt(msg).split(",")]

    if user_response == ["s"]:
        return []

    if user_response == ["q"]:
        if click.confirm(
            "Are you sure you want to quit annotating the dataset? (progress will be saved)"
        ):
            raise AnnotationAbortError
        return get_data_categories_annotation(dataset_member, valid_categories)

    if validate:
        try:
            validate_data_categories(user_response, valid_categories)
        except (FidesValidationError, ValueError):
            click.secho(
                f"[{user_response}] is not a valid data category, please re-confirm and try again!",
                fg="red",
            )
            user_response = get_data_categories_annotation(
                dataset_member, valid_categories
            )

    return user_response


def annotate_dataset(
    config: FidesConfig,
    dataset_file: str,
    resource_type: str = "data_category",
    annotate_all: bool = False,
    validate: bool = True,
    include_null: bool = False,
) -> None:
    """
    Given a dataset.yml-like file, walk the user through an interactive cli to provide data categories
    for members of the dataset that do not have any specified
    Args:
        dataset_file: the file name for the dataset to annotate
        resource_type: the type of data resource to point to for assistance (via visualization web page)
        annotate_all: flag to annotate all members of a dataset (default False: only annotate fields)
        validate: flag to check user inputs for formatting and data category presence
        include_null: flag to write out all empty attributes with null (default False: only write populated attributes)

    Returns:
        Write the amended dataset file in place
    """
    output_dataset = []

    datasets = ingest_manifests(dataset_file)["dataset"]
    resources = api_helpers.list_server_resources(
        url=str(config.cli.server_url),
        resource_type=resource_type,
        headers=config.user.auth_header
        or {"Authorization": "Bearer fakekey"},  # Alternate header required for mypy
        exclude_keys=[],
    )

    if not resources:
        click.secho(
            "No server resources were found.",
            fg="red",
        )
        return

    existing_categories: List[str] = [
        resource.fides_key
        if isinstance(resource, FidesModel)
        else resource["fides_key"]
        for resource in resources
    ]

    for dataset in datasets:
        current_dataset = Dataset(**dataset)
        try:
            click.secho(f"\n####\nAnnotating Dataset: [{current_dataset.name}]")

            # Check for data_categories at the Dataset level
            if annotate_all and not dataset.data_categories:
                click.secho(f"Database [{dataset.name}] has no data categories")
                dataset.data_categories = get_data_categories_annotation(
                    dataset_member=dataset,
                    valid_categories=existing_categories,
                    validate=validate,
                )

            annotate_collections(
                current_dataset,
                existing_categories,
                annotate_all,
                validate,
            )

            if include_null:
                output_dataset.append(current_dataset.dict())
            else:
                output_dataset.append(current_dataset.dict(exclude_none=True))
        except AnnotationAbortError:
            if include_null:
                output_dataset.append(current_dataset.dict())
            else:
                output_dataset.append(current_dataset.dict(exclude_none=True))
            break
    manifests.write_manifest(dataset_file, output_dataset, "dataset")
    echo_green("Annotation process complete.")


def annotate_collections(
    current_dataset: Dataset,
    existing_categories: List[str],
    annotate_all: bool,
    validate: bool,
) -> None:
    """
    Check for data_categories at the collections level
    """

    for table in current_dataset.collections:
        click.secho(f"####\nAnnotating Table: [{table.name}]\n")
        if annotate_all and not table.data_categories:
            click.secho(f"Table [{table.name}] has no data categories")
            table.data_categories = get_data_categories_annotation(
                dataset_member=table,
                valid_categories=existing_categories,
                validate=validate,
            )

        annotate_fields(table, existing_categories, validate)


def annotate_fields(
    table: DatasetCollection,
    existing_categories: List[str],
    validate: bool,
) -> None:
    """
    Check for data_categories at the field level
    """

    for field in get_all_level_fields(table.fields):
        if not field.data_categories:
            click.secho(
                f"Field [{table.name}.{field.name}] has no data categories\n",
                fg="red",
            )

            user_categories = get_data_categories_annotation(
                dataset_member=field,
                valid_categories=existing_categories,
                validate=validate,
            )

            click.secho(
                f"""Setting data categories for {table.name}.{field.name} to:
            {user_categories}\n""",
                fg="green",
            )

            field.data_categories = user_categories



from typing import Generator

from fastapi import Depends, Security
from fastapi.security import SecurityScopes
from sqlalchemy.orm import Session

from fides.core.config import get_config
from fides.lib.db.session import get_db_session
from fides.lib.models.client import ClientDetail
from fides.lib.oauth.api.urn_registry import TOKEN, V1_URL_PREFIX
from fides.lib.oauth.oauth_util import verify_oauth_client as verify
from fides.lib.oauth.schemas.oauth import OAuth2ClientCredentialsBearer


def get_db() -> Generator[Session, None, None]:
    """Return our database session.

    This should be overridden by the installing package.
    """
    try:
        SessionLocal = get_db_session(get_config())
        db = SessionLocal()
        yield db
    finally:
        db.close()


def oauth2_scheme() -> OAuth2ClientCredentialsBearer:
    """Creates the oauth2 scheme from the token.

    This should be overridden by the installing package.
    """
    return OAuth2ClientCredentialsBearer(tokenUrl=f"{V1_URL_PREFIX}{TOKEN}")


def verify_oauth_client(
    security_scopes: SecurityScopes,
    authorization: str = Security(oauth2_scheme()),
    db: Session = Depends(get_db),
) -> ClientDetail:
    """Calls oauth_util.verify_oauth_client.

    This is here because config values are needed, this dependency should be overridden
    by the installing library.
    """
    config = get_config()
    return verify(security_scopes, authorization, db=db, config=config)



"""final unified ctl merge

Revision ID: fb6b0150d6e4
Revises: 3d9c476d7cea, 6b287ff8cde4
Create Date: 2022-10-06 14:56:13.267477

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "fb6b0150d6e4"
down_revision = ("3d9c476d7cea", "6b287ff8cde4")
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



# Prefix
V1_URL_PREFIX = "/api/v1"

# Login URLs
LOGIN = "/login"

# Oauth Client URLs
TOKEN = "/oauth/token"

# User URLs
USERS = "/user"
USER_DETAIL = "/user/{user_id}"
USER_PASSWORD_RESET = "/user/{user_id}/reset-password"

# User Permission URLs
USER_PERMISSIONS = "/user/{user_id}/permission"



"""manual webhooks

Revision ID: 7a4f4042091e
Revises: d8df7ff7aab4
Create Date: 2022-09-07 14:10:28.106679

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "7a4f4042091e"
down_revision = "d8df7ff7aab4"
branch_labels = None
depends_on = None


def upgrade():
    # Add new table: AccessManualWebhook
    op.create_table(
        "accessmanualwebhook",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("connection_config_id", sa.String(), nullable=False),
        sa.Column("fields", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.ForeignKeyConstraint(
            ["connection_config_id"],
            ["connectionconfig.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("connection_config_id"),
    )
    op.create_index(
        op.f("ix_accessmanualwebhook_id"), "accessmanualwebhook", ["id"], unique=False
    )

    # Add to ConnectionType enum
    op.execute("alter type connectiontype rename to connectiontype_old")
    op.execute(
        "create type connectiontype as enum('postgres', 'mongodb', 'mysql', 'https', 'snowflake', 'redshift', 'mssql', 'mariadb', 'bigquery', 'saas', 'manual', 'email', 'manual_webhook')"
    )
    op.execute(
        (
            "alter table connectionconfig alter column connection_type type connectiontype using "
            "connection_type::text::connectiontype"
        )
    )
    op.execute("drop type connectiontype_old")

    # Add to PrivacyRequestStatus enum
    op.execute("alter type privacyrequeststatus add value 'requires_input'")


def downgrade():
    # Remove AccessManualWebhook table
    op.drop_index(op.f("ix_accessmanualwebhook_id"), table_name="accessmanualwebhook")
    op.drop_table("accessmanualwebhook")

    # Remove 'manual_webhook' from ConnectionType
    op.execute("alter type connectiontype rename to connectiontype_old")
    op.execute(
        "create type connectiontype as enum('postgres', 'mongodb', 'mysql', 'https', 'snowflake', 'redshift', 'mssql', 'mariadb', 'bigquery', 'saas', 'manual', 'email')"
    )
    op.execute(
        (
            "alter table connectionconfig alter column connection_type type connectiontype using "
            "connection_type::text::connectiontype"
        )
    )
    op.execute("drop type connectiontype_old")

    # Remove PrivacyRequestStatus enum: requires_input
    op.execute("alter type privacyrequeststatus rename to privacyrequeststatus_old")
    op.execute(
        "create type privacyrequeststatus as enum('in_processing', 'complete', 'pending', 'error', 'paused', 'approved', 'denied', 'canceled', 'identity_unverified')"
    )
    op.execute(
        (
            "alter table privacyrequest alter column status type privacyrequeststatus using "
            "status::text::privacyrequeststatus"
        )
    )
    op.execute("drop type privacyrequeststatus_old")



from __future__ import annotations

import json
from typing import Any, Callable, Dict, List, Optional, Union

import requests
import sendgrid
from loguru import logger
from sendgrid.helpers.mail import Content, Email, Mail, To
from sqlalchemy.orm import Session
from twilio.base.exceptions import TwilioRestException
from twilio.rest import Client

from fides.api.ops.common_exceptions import MessageDispatchException
from fides.api.ops.email_templates import get_email_template
from fides.api.ops.models.messaging import (  # type: ignore[attr-defined]
    EMAIL_MESSAGING_SERVICES,
    MessagingConfig,
    get_messaging_method,
)
from fides.api.ops.models.privacy_request import (
    CheckpointActionRequired,
    PrivacyRequestError,
    PrivacyRequestNotifications,
)
from fides.api.ops.schemas.messaging.messaging import (
    AccessRequestCompleteBodyParams,
    EmailForActionType,
    ErrorNotificaitonBodyParams,
    FidesopsMessage,
    MessagingActionType,
    MessagingMethod,
    MessagingServiceDetails,
    MessagingServiceSecrets,
    MessagingServiceType,
    RequestReceiptBodyParams,
    RequestReviewDenyBodyParams,
    SubjectIdentityVerificationBodyParams,
)
from fides.api.ops.schemas.redis_cache import Identity
from fides.api.ops.tasks import MESSAGING_QUEUE_NAME, DatabaseTask, celery_app
from fides.api.ops.util.logger import Pii
from fides.core.config import get_config

CONFIG = get_config()


EMAIL_JOIN_STRING = ", "


def check_and_dispatch_error_notifications(db: Session) -> None:
    privacy_request_notifications = PrivacyRequestNotifications.all(db=db)
    if not privacy_request_notifications:
        return None

    unsent_errors = PrivacyRequestError.filter(
        db=db, conditions=(PrivacyRequestError.message_sent.is_(False))
    ).all()
    if not unsent_errors:
        return None

    email_config = (
        CONFIG.notifications.notification_service_type in EMAIL_MESSAGING_SERVICES
    )

    if (
        email_config
        and len(unsent_errors) >= privacy_request_notifications[0].notify_after_failures
    ):
        for email in privacy_request_notifications[0].email.split(EMAIL_JOIN_STRING):
            dispatch_message_task.apply_async(
                queue=MESSAGING_QUEUE_NAME,
                kwargs={
                    "message_meta": FidesopsMessage(
                        action_type=MessagingActionType.PRIVACY_REQUEST_ERROR_NOTIFICATION,
                        body_params=ErrorNotificaitonBodyParams(
                            unsent_errors=len(unsent_errors)
                        ),
                    ).dict(),
                    "service_type": CONFIG.notifications.notification_service_type,
                    "to_identity": {"email": email},
                },
            )

        for error in unsent_errors:
            error.update(db=db, data={"message_sent": True})

    return None


@celery_app.task(base=DatabaseTask, bind=True)
def dispatch_message_task(
    self: DatabaseTask,
    message_meta: Dict[str, Any],
    service_type: Optional[str],
    to_identity: Dict[str, Any],
) -> None:
    """
    A wrapper function to dispatch a message task into the Celery queues
    """
    schema = FidesopsMessage.parse_obj(message_meta)
    with self.get_new_session() as db:
        dispatch_message(
            db,
            schema.action_type,
            Identity.parse_obj(to_identity),
            service_type,
            schema.body_params,
        )


def dispatch_message(
    db: Session,
    action_type: MessagingActionType,
    to_identity: Optional[Identity],
    service_type: Optional[str],
    message_body_params: Optional[
        Union[
            AccessRequestCompleteBodyParams,
            SubjectIdentityVerificationBodyParams,
            RequestReceiptBodyParams,
            RequestReviewDenyBodyParams,
            List[CheckpointActionRequired],
        ]
    ] = None,
) -> None:
    """
    Sends a message to `to_identity` with content supplied in `message_body_params`
    """
    if not to_identity:
        logger.error("Message failed to send. No identity supplied.")
        raise MessageDispatchException("No identity supplied.")
    if not service_type:
        logger.error("Message failed to send. No notification service type configured.")
        raise MessageDispatchException("No notification service type configured.")

    logger.info("Retrieving message config")
    messaging_config: MessagingConfig = MessagingConfig.get_configuration(
        db=db, service_type=service_type
    )
    logger.info(
        "Building appropriate message template for action type: {}", action_type
    )
    messaging_method = get_messaging_method(service_type)
    message: Optional[Union[EmailForActionType, str]] = None
    if messaging_method == MessagingMethod.EMAIL:
        message = _build_email(
            action_type=action_type,
            body_params=message_body_params,
        )
    elif messaging_method == MessagingMethod.SMS:
        message = _build_sms(
            action_type=action_type,
            body_params=message_body_params,
        )
    else:
        logger.error(
            "Notification service type is not valid: {}",
            CONFIG.notifications.notification_service_type,
        )
        raise MessageDispatchException(
            f"Notification service type is not valid: {CONFIG.notifications.notification_service_type}"
        )
    messaging_service: MessagingServiceType = messaging_config.service_type  # type: ignore
    logger.info(
        "Retrieving appropriate dispatcher for email service: {}", messaging_service
    )
    dispatcher: Optional[
        Callable[[MessagingConfig, Any, Optional[str]], None]
    ] = _get_dispatcher_from_config_type(message_service_type=messaging_service)
    if not dispatcher:
        logger.error(
            "Dispatcher has not been implemented for message service type: {}",
            messaging_service,
        )
        raise MessageDispatchException(
            f"Dispatcher has not been implemented for message service type: {messaging_service}"
        )
    logger.info(
        "Starting message dispatch for messaging service with action type: {}",
        action_type,
    )
    dispatcher(
        messaging_config,
        message,
        to_identity.email
        if messaging_method == MessagingMethod.EMAIL
        else to_identity.phone_number,
    )


def _build_sms(  # pylint: disable=too-many-return-statements
    action_type: MessagingActionType,
    body_params: Any,
) -> str:
    separator = ","
    if action_type == MessagingActionType.SUBJECT_IDENTITY_VERIFICATION:
        return (
            f"Your privacy request verification code is {body_params.verification_code}. "
            f"Please return to the Privacy Center and enter the code to continue. "
            f"This code will expire in {body_params.get_verification_code_ttl_minutes()} minutes"
        )
    if action_type == MessagingActionType.CONSENT_REQUEST:
        return (
            f"Your consent request verification code is {body_params.verification_code}. "
            "Please return to the consent request page and enter the code to continue. "
            f"This code will expire in {body_params.get_verification_code_ttl_minutes()} minutes"
        )
    if action_type == MessagingActionType.PRIVACY_REQUEST_RECEIPT:
        if len(body_params.request_types) > 1:
            return f"The following requests have been received: {separator.join(body_params.request_types)}"
        return f"Your {body_params.request_types[0]} request has been received"
    if action_type == MessagingActionType.PRIVACY_REQUEST_COMPLETE_ACCESS:
        if len(body_params.download_links) > 1:
            return (
                "Your data access has been completed and can be downloaded at the following links. "
                "For security purposes, these secret links will expire in 24 hours: "
                f"{separator.join(body_params.download_links)}"
            )
        return (
            f"Your data access has been completed and can be downloaded at {body_params.download_links[0]}. "
            f"For security purposes, this secret link will expire in 24 hours."
        )
    if action_type == MessagingActionType.PRIVACY_REQUEST_COMPLETE_DELETION:
        return "Your privacy request for deletion has been completed."
    if action_type == MessagingActionType.PRIVACY_REQUEST_REVIEW_APPROVE:
        return "Your privacy request has been approved and is currently processing."
    if action_type == MessagingActionType.PRIVACY_REQUEST_REVIEW_DENY:
        if body_params.rejection_reason:
            return f"Your privacy request has been denied for the following reason: {body_params.rejection_reason}"
        return "Your privacy request has been denied."
    logger.error("Message action type {} is not implemented", action_type)
    raise MessageDispatchException(
        f"Message action type {action_type} is not implemented"
    )


def _build_email(  # pylint: disable=too-many-return-statements
    action_type: MessagingActionType,
    body_params: Any,
) -> EmailForActionType:
    if action_type == MessagingActionType.CONSENT_REQUEST:
        template = get_email_template(action_type)
        return EmailForActionType(
            subject="Your one-time code",
            body=template.render(
                {
                    "code": body_params.verification_code,
                    "minutes": body_params.get_verification_code_ttl_minutes(),
                }
            ),
        )
    if action_type == MessagingActionType.SUBJECT_IDENTITY_VERIFICATION:
        template = get_email_template(action_type)
        return EmailForActionType(
            subject="Your one-time code",
            body=template.render(
                {
                    "code": body_params.verification_code,
                    "minutes": body_params.get_verification_code_ttl_minutes(),
                }
            ),
        )
    if action_type == MessagingActionType.MESSAGE_ERASURE_REQUEST_FULFILLMENT:
        base_template = get_email_template(action_type)
        return EmailForActionType(
            subject="Data erasure request",
            body=base_template.render(
                {"dataset_collection_action_required": body_params}
            ),
        )
    if action_type == MessagingActionType.PRIVACY_REQUEST_RECEIPT:
        base_template = get_email_template(action_type)
        return EmailForActionType(
            subject="Your request has been received",
            body=base_template.render({"request_types": body_params.request_types}),
        )
    if action_type == MessagingActionType.PRIVACY_REQUEST_COMPLETE_ACCESS:
        base_template = get_email_template(action_type)
        return EmailForActionType(
            subject="Your data is ready to be downloaded",
            body=base_template.render(
                {
                    "download_links": body_params.download_links,
                }
            ),
        )
    if action_type == MessagingActionType.PRIVACY_REQUEST_COMPLETE_DELETION:
        base_template = get_email_template(action_type)
        return EmailForActionType(
            subject="Your data has been deleted",
            body=base_template.render(),
        )
    if action_type == MessagingActionType.PRIVACY_REQUEST_ERROR_NOTIFICATION:
        base_template = get_email_template(action_type)
        return EmailForActionType(
            subject="Privacy Request Error Alert",
            body=base_template.render(),
        )
    if action_type == MessagingActionType.PRIVACY_REQUEST_REVIEW_APPROVE:
        base_template = get_email_template(action_type)
        return EmailForActionType(
            subject="Your request has been approved",
            body=base_template.render(),
        )
    if action_type == MessagingActionType.PRIVACY_REQUEST_REVIEW_DENY:
        base_template = get_email_template(action_type)
        return EmailForActionType(
            subject="Your request has been denied",
            body=base_template.render(
                {"rejection_reason": body_params.rejection_reason}
            ),
        )
    logger.error("Message action type {} is not implemented", action_type)
    raise MessageDispatchException(
        f"Message action type {action_type} is not implemented"
    )


def _get_dispatcher_from_config_type(
    message_service_type: MessagingServiceType,
) -> Optional[Callable[[MessagingConfig, Any, Optional[str]], None]]:
    """Determines which dispatcher to use based on message service type"""
    if message_service_type == MessagingServiceType.MAILGUN:
        return _mailgun_dispatcher
    if message_service_type == MessagingServiceType.TWILIO_TEXT:
        return _twilio_sms_dispatcher
    if message_service_type == MessagingServiceType.TWILIO_EMAIL:
        return _twilio_email_dispatcher
    return None


def _mailgun_dispatcher(
    messaging_config: MessagingConfig,
    message: EmailForActionType,
    to: Optional[str],
) -> None:
    """Dispatches email using mailgun"""
    if not to:
        logger.error("Message failed to send. No email identity supplied.")
        raise MessageDispatchException("No email identity supplied.")
    if not messaging_config.details or not messaging_config.secrets:
        logger.error(
            "Message failed to send. No mailgun config details or secrets supplied."
        )
        raise MessageDispatchException("No mailgun config details or secrets supplied.")
    base_url = (
        "https://api.mailgun.net"
        if messaging_config.details[MessagingServiceDetails.IS_EU_DOMAIN.value] is False
        else "https://api.eu.mailgun.net"
    )

    domain = messaging_config.details[MessagingServiceDetails.DOMAIN.value]

    try:
        # Check if a fides template exists
        template_test = requests.get(
            f"{base_url}/{messaging_config.details[MessagingServiceDetails.API_VERSION.value]}/{domain}/templates/fides",
            auth=(
                "api",
                messaging_config.secrets[MessagingServiceSecrets.MAILGUN_API_KEY.value],
            ),
        )

        data = {
            "from": f"<mailgun@{domain}>",
            "to": [to.strip()],
            "subject": message.subject,
        }

        if template_test.status_code == 200:
            data["template"] = "fides"
            data["h:X-Mailgun-Variables"] = json.dumps(
                {"fides_email_body": message.body}
            )
            response = requests.post(
                f"{base_url}/{messaging_config.details[MessagingServiceDetails.API_VERSION.value]}/{domain}/messages",
                auth=(
                    "api",
                    messaging_config.secrets[
                        MessagingServiceSecrets.MAILGUN_API_KEY.value
                    ],
                ),
                data=data,
            )

            if not response.ok:
                logger.error(
                    "Email failed to send with status code: %s", response.status_code
                )
                raise MessageDispatchException(
                    f"Email failed to send with status code {response.status_code}"
                )
        else:
            data["html"] = message.body
            response = requests.post(
                f"{base_url}/{messaging_config.details[MessagingServiceDetails.API_VERSION.value]}/{domain}/messages",
                auth=(
                    "api",
                    messaging_config.secrets[
                        MessagingServiceSecrets.MAILGUN_API_KEY.value
                    ],
                ),
                data=data,
            )
            if not response.ok:
                logger.error(
                    "Email failed to send with status code: %s", response.status_code
                )
                raise MessageDispatchException(
                    f"Email failed to send with status code {response.status_code}"
                )
    except Exception as e:
        logger.error("Email failed to send: {}", Pii(str(e)))
        raise MessageDispatchException(f"Email failed to send due to: {Pii(e)}")


def _twilio_email_dispatcher(
    messaging_config: MessagingConfig,
    message: EmailForActionType,
    to: Optional[str],
) -> None:
    """Dispatches email using twilio sendgrid"""
    if not to:
        logger.error("Message failed to send. No email identity supplied.")
        raise MessageDispatchException("No email identity supplied.")
    if not messaging_config.details or not messaging_config.secrets:
        logger.error(
            "Message failed to send. No twilio email config details or secrets supplied."
        )
        raise MessageDispatchException(
            "No twilio email config details or secrets supplied."
        )

    try:
        sg = sendgrid.SendGridAPIClient(
            api_key=messaging_config.secrets[
                MessagingServiceSecrets.TWILIO_API_KEY.value
            ]
        )
        from_email = Email(
            messaging_config.details[MessagingServiceDetails.TWILIO_EMAIL_FROM.value]
        )
        to_email = To(to.strip())
        subject = message.subject
        content = Content("text/html", message.body)
        mail = Mail(from_email, to_email, subject, content)
        response = sg.client.mail.send.post(request_body=mail.get())
        if response.status_code >= 400:
            logger.error(
                "Email failed to send: %s: %s",
                response.status_code,
                Pii(str(response.body)),
            )
            raise MessageDispatchException(
                f"Email failed to send: {response.status_code}, {Pii(str(response.body))}"
            )
    except Exception as e:
        logger.error("Email failed to send: {}", Pii(str(e)))
        raise MessageDispatchException(f"Email failed to send due to: {Pii(e)}")


def _twilio_sms_dispatcher(
    messaging_config: MessagingConfig,
    message: str,
    to: Optional[str],
) -> None:
    """Dispatches SMS using Twilio"""
    if not to:
        logger.error("Message failed to send. No phone identity supplied.")
        raise MessageDispatchException("No phone identity supplied.")
    if messaging_config.secrets is None:
        logger.error("Message failed to send. No config secrets supplied.")
        raise MessageDispatchException("No config secrets supplied.")
    account_sid = messaging_config.secrets[
        MessagingServiceSecrets.TWILIO_ACCOUNT_SID.value
    ]
    auth_token = messaging_config.secrets[
        MessagingServiceSecrets.TWILIO_AUTH_TOKEN.value
    ]
    messaging_service_id = messaging_config.secrets[
        MessagingServiceSecrets.TWILIO_MESSAGING_SERVICE_SID.value
    ]
    sender_phone_number = messaging_config.secrets[
        MessagingServiceSecrets.TWILIO_SENDER_PHONE_NUMBER.value
    ]

    client = Client(account_sid, auth_token)
    try:
        if messaging_service_id:
            client.messages.create(
                to=to, messaging_service_sid=messaging_service_id, body=message
            )
        elif sender_phone_number:
            client.messages.create(to=to, from_=sender_phone_number, body=message)
        else:
            logger.error(
                "Message failed to send. Either sender phone number or messaging service sid must be provided."
            )
            raise MessageDispatchException(
                "Message failed to send. Either sender phone number or messaging service sid must be provided."
            )
    except TwilioRestException as e:
        logger.error("Twilio SMS failed to send: {}", Pii(str(e)))
        raise MessageDispatchException(f"Twilio SMS failed to send due to: {Pii(e)}")



from typing import Optional

from pydantic import validator

from .fides_settings import FidesSettings

ENV_PREFIX = "FIDES__NOTIFICATIONS__"


class NotificationSettings(FidesSettings):
    """Configuration settings for data subject and/or data processor notifications"""

    send_request_completion_notification: bool = False
    send_request_receipt_notification: bool = False
    send_request_review_notification: bool = False
    notification_service_type: Optional[str] = None

    @validator("notification_service_type", pre=True)
    @classmethod
    def validate_notification_service_type(cls, value: Optional[str]) -> Optional[str]:
        """Ensure the provided type is a valid value."""
        if value:
            valid_values = ["MAILGUN", "TWILIO_TEXT", "TWILIO_EMAIL"]
            value = value.upper()  # force uppercase for safety

            if value not in valid_values:
                raise ValueError(
                    f"Invalid NOTIFICATION_SERVICE_TYPE provided '{value}', must be one of: {', '.join([level for level in valid_values])}"
                )

        return value

    class Config:
        env_prefix = ENV_PREFIX



from typing import Dict, List

from fideslang import FidesModelType
from sqlalchemy.ext.asyncio import AsyncSession

from fides.api.ctl.database.crud import get_resource, list_resource
from fides.api.ctl.sql_models import (  # type: ignore[attr-defined]
    models_with_default_field,
)
from fides.api.ctl.utils import errors
from fides.api.ctl.utils.api_router import APIRouter
from fides.lib.db.base import Base  # type: ignore[attr-defined]

API_PREFIX = "/api/v1"


def get_resource_type(router: APIRouter) -> str:
    """
    Get the resource type from the prefix of an API router
    Args:
        router: Api router from which to extract the resource type

    Returns:
        The router's resource type
    """
    return router.prefix.replace(f"{API_PREFIX}/", "", 1)


async def forbid_if_editing_is_default(
    sql_model: Base,
    fides_key: str,
    payload: FidesModelType,
    async_session: AsyncSession,
) -> None:
    """
    Raise a forbidden error if the user is trying modify the `is_default` field
    """
    if sql_model in models_with_default_field:
        resource = await get_resource(sql_model, fides_key, async_session)
        if resource.is_default != payload.is_default:
            raise errors.ForbiddenError(sql_model.__name__, fides_key)


async def forbid_if_default(
    sql_model: Base, fides_key: str, async_session: AsyncSession
) -> None:
    """
    Raise a forbidden error if the user is trying to operate on a resource
    with `is_default=True`
    """
    if sql_model in models_with_default_field:
        resource = await get_resource(sql_model, fides_key, async_session)
        if resource.is_default:
            raise errors.ForbiddenError(sql_model.__name__, fides_key)


async def forbid_if_editing_any_is_default(
    sql_model: Base, resources: List[Dict], async_session: AsyncSession
) -> None:
    """
    Raise a forbidden error if any of the existing resources' `is_default`
    field is being modified, or if there is a new resource with `is_default=True`
    """
    if sql_model in models_with_default_field:
        fides_keys = [resource["fides_key"] for resource in resources]
        existing_resources = {
            r.fides_key: r
            for r in await list_resource(sql_model, async_session)
            if r.fides_key in fides_keys
        }
        for resource in resources:
            if existing_resources.get(resource["fides_key"]) is None:
                # new resource is being upserted
                if resource["is_default"]:
                    raise errors.ForbiddenError(
                        sql_model.__name__, resource["fides_key"]
                    )
            elif (
                resource["is_default"]
                != existing_resources[resource["fides_key"]].is_default
            ):
                raise errors.ForbiddenError(sql_model.__name__, resource["fides_key"])



"""Utils to help with API calls."""
import glob
import re
from functools import partial
from hashlib import sha1
from json.decoder import JSONDecodeError
from os.path import isfile
from typing import Iterator, List

import click
import requests
import sqlalchemy
from fideslang.models import DatasetField, FidesModel
from loguru import logger
from pydantic import ValidationError
from sqlalchemy.engine import Engine
from sqlalchemy.exc import SQLAlchemyError

from fides.connectors.models import ConnectorAuthFailureException

logger.bind(name="server_api")

echo_red = partial(click.secho, fg="red", bold=True)
echo_green = partial(click.secho, fg="green", bold=True)


def check_response_auth(response: requests.Response) -> requests.Response:
    """
    Verify that a response object is 'ok', otherwise print the error and raise
    an exception.
    """
    if response.status_code in [401, 403]:
        echo_red("Authorization Error: please try 'fides user login' and try again.")
        raise SystemExit(1)
    return response


def check_response(response: requests.Response) -> requests.Response:
    """
    Check that a response has valid JSON.
    """

    try:
        response.json()
    except JSONDecodeError as json_error:
        logger.error(response.status_code)
        logger.error(response.text)
        raise json_error
    else:
        return response


def validate_db_engine(connection_string: str) -> None:
    """
    Use SQLAlchemy to create a DB engine.
    """
    try:
        engine = sqlalchemy.create_engine(connection_string)
        with engine.begin() as connection:
            connection.execute("SELECT 1")
    except SQLAlchemyError as error:
        raise ConnectorAuthFailureException(error)


def get_db_engine(connection_string: str) -> Engine:
    """
    Use SQLAlchemy to create a DB engine.
    """
    try:
        engine = sqlalchemy.create_engine(
            connection_string, connect_args={"connect_timeout": 10}
        )
    except Exception as err:
        raise Exception("Failed to create engine!") from err

    try:
        with engine.begin() as connection:
            connection.execute("SELECT 1")
    except Exception as err:
        raise Exception(f"Database connection failed with engine:\n{engine}!") from err
    return engine


def get_all_level_fields(fields: list) -> Iterator[DatasetField]:
    """
    Traverses all levels of fields that exist in a dataset
    returning them for individual evaluation.
    """
    for field in fields:
        yield field
        if field.fields:
            for nested_field in get_all_level_fields(field.fields):
                yield nested_field


def get_manifest_list(manifests_dir: str) -> List[str]:
    """Get a list of manifest files from the manifest directory."""
    yml_endings = ["yml", "yaml"]

    if isfile(manifests_dir) and manifests_dir.split(".")[-1] in yml_endings:
        return [manifests_dir]

    manifest_list = []
    for yml_ending in yml_endings:
        manifest_list += glob.glob(f"{manifests_dir}/**/*.{yml_ending}", recursive=True)

    return manifest_list


def check_fides_key(proposed_fides_key: str) -> str:
    """
    A helper function to automatically sanitize
    an invalid FidesKey.
    """
    try:
        FidesModel(fides_key=proposed_fides_key)
        return proposed_fides_key
    except ValidationError as error:
        echo_red(error)
        return sanitize_fides_key(proposed_fides_key)


def sanitize_fides_key(proposed_fides_key: str) -> str:
    """
    Attempts to manually sanitize a fides key if restricted
    characters are discovered.
    """
    sanitized_fides_key = re.sub(r"[^a-zA-Z0-9_.-]", "_", proposed_fides_key)
    return sanitized_fides_key


def generate_unique_fides_key(
    proposed_fides_key: str, database_host: str, database_name: str
) -> str:
    """
    Uses host and name to generate a UUID to be
    appended to the fides_key of a dataset
    """
    fides_key_uuid = sha1(
        "-".join([database_host, database_name, proposed_fides_key]).encode()
    )
    return f"{proposed_fides_key}_{fides_key_uuid.hexdigest()[:10]}"


def git_is_dirty(dir_to_check: str = ".") -> bool:
    """
    Checks to see if the local repo has unstaged changes.
    Can also specify a directory to check.
    """

    try:
        from git.repo import Repo
        from git.repo.fun import is_git_dir
    except ImportError:
        print("Git executable not detected, skipping git check...")
        return False

    git_dir_path = ".git/"
    if not is_git_dir(git_dir_path):
        print(f"No git repo detected at '{git_dir_path}', skipping git check...")
        return False

    repo = Repo()
    git_session = repo.git()

    dirty_phrases = ["Changes not staged for commit:", "Untracked files:"]
    git_status = git_session.status(dir_to_check).split("\n")
    is_dirty = any(phrase in git_status for phrase in dirty_phrases)
    return is_dirty



"""add data_responsibility_title attribute for system

Revision ID: fdd28c39a679
Revises: 7a94527643a1
Create Date: 2022-03-21 17:37:23.855761

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "fdd28c39a679"
down_revision = "7a94527643a1"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("systems", sa.Column("data_responsibility_title", sa.String()))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("systems", "data_responsibility_title")
    # ### end Alembic commands ###



import secrets
from typing import Dict, List, Optional, TypeVar

from loguru import logger

from fides.api.ops.schemas.masking.masking_secrets import (
    MaskingSecretCache,
    MaskingSecretMeta,
    SecretType,
)
from fides.api.ops.util.cache import get_cache, get_masking_secret_cache_key

T = TypeVar("T")


class SecretsUtil:
    @staticmethod
    def get_or_generate_secret(
        privacy_request_id: Optional[str],
        secret_type: SecretType,
        masking_secret_meta: MaskingSecretMeta[T],
    ) -> Optional[T]:
        if privacy_request_id is not None:
            secret = SecretsUtil._get_secret_from_cache(
                privacy_request_id, secret_type, masking_secret_meta
            )
            if not secret:
                logger.warning(
                    "Secret type {} expected from cache but was not present for masking strategy {}",
                    secret_type,
                    masking_secret_meta.masking_strategy,
                )
            return secret

        # expected for standalone masking service
        return masking_secret_meta.generate_secret_func(
            masking_secret_meta.secret_length
        )

    @staticmethod
    def _get_secret_from_cache(
        privacy_request_id: str,
        secret_type: SecretType,
        masking_secret_meta: MaskingSecretMeta[T],
    ) -> Optional[T]:
        cache = get_cache()
        masking_secret_cache_key: str = get_masking_secret_cache_key(
            privacy_request_id=privacy_request_id,
            masking_strategy=masking_secret_meta.masking_strategy,
            secret_type=secret_type,
        )
        return cache.get_encoded_by_key(masking_secret_cache_key)

    @staticmethod
    def generate_secret_string(length: int) -> str:
        return secrets.token_urlsafe(length)

    @staticmethod
    def generate_secret_bytes(length: int) -> bytes:
        return secrets.token_bytes(length)

    @staticmethod
    def build_masking_secrets_for_cache(
        masking_secret_meta: Dict[SecretType, MaskingSecretMeta[T]],
    ) -> List[MaskingSecretCache[T]]:
        masking_secrets = []
        for secret_type in masking_secret_meta.keys():
            meta = masking_secret_meta[secret_type]
            secret: T = meta.generate_secret_func(meta.secret_length)
            masking_secrets.append(
                MaskingSecretCache[T](  # type: ignore
                    secret=secret,
                    masking_strategy=meta.masking_strategy,
                    secret_type=secret_type,
                )
            )
        return masking_secrets



from typing import Callable, List

from fastapi import Request
from fastapi.responses import JSONResponse
from starlette.status import HTTP_500_INTERNAL_SERVER_ERROR

from fides.api.ops.common_exceptions import FunctionalityNotConfigured


class ExceptionHandlers:
    @staticmethod
    def functionality_not_configured_handler(
        request: Request, exc: FunctionalityNotConfigured
    ) -> JSONResponse:
        return JSONResponse(
            status_code=HTTP_500_INTERNAL_SERVER_ERROR, content={"message": str(exc)}
        )

    @classmethod
    def get_handlers(
        cls,
    ) -> List[Callable[[Request, FunctionalityNotConfigured], JSONResponse]]:
        return [ExceptionHandlers.functionality_not_configured_handler]



"""prepend tables with ctl

Revision ID: aaa81d97a6f6
Revises: f53e04e5b7f5
Create Date: 2022-08-11 09:59:38.709501

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "aaa81d97a6f6"
down_revision = "f53e04e5b7f5"
branch_labels = None
depends_on = None


def upgrade():
    """Rename tables to avoid collisions with fidesops."""

    # Data Categories
    op.rename_table("data_categories", "ctl_data_categories")
    op.execute("ALTER INDEX data_categories_pkey RENAME TO ctl_data_categories_pkey")
    op.execute(
        "ALTER INDEX ix_data_categories_fides_key RENAME TO ix_ctl_data_categories_fides_key"
    )
    op.execute("ALTER INDEX ix_data_categories_id RENAME TO ix_ctl_data_categories_id")

    # Data Subjects
    op.rename_table("data_subjects", "ctl_data_subjects")
    op.execute("ALTER INDEX data_subjects_pkey RENAME TO ctl_data_subjects_pkey")
    op.execute(
        "ALTER INDEX ix_data_subjects_fides_key RENAME TO ix_ctl_data_subjects_fides_key"
    )
    op.execute("ALTER INDEX ix_data_subjects_id RENAME TO ix_ctl_data_subjects_id")

    # Data Uses
    op.rename_table("data_uses", "ctl_data_uses")
    op.execute("ALTER INDEX data_uses_pkey RENAME TO ctl_data_uses_pkey")
    op.execute(
        "ALTER INDEX ix_data_uses_fides_key RENAME TO ix_ctl_data_uses_fides_key"
    )
    op.execute("ALTER INDEX ix_data_uses_id RENAME TO ix_ctl_data_uses_id")

    # Data Qualifiers
    op.rename_table("data_qualifiers", "ctl_data_qualifiers")
    op.execute("ALTER INDEX data_qualifiers_pkey RENAME TO ctl_data_qualifiers_pkey")
    op.execute(
        "ALTER INDEX ix_data_qualifiers_fides_key RENAME TO ix_ctl_data_qualifiers_fides_key"
    )
    op.execute("ALTER INDEX ix_data_qualifiers_id RENAME TO ix_ctl_data_qualifiers_id")

    # Datasets
    op.rename_table("datasets", "ctl_datasets")
    op.execute("ALTER INDEX data_sets_pkey RENAME TO ctl_datasets_pkey")
    op.execute("ALTER INDEX ix_datasets_fides_key RENAME TO ix_ctl_datasets_fides_key")
    op.execute("ALTER INDEX ix_datasets_id RENAME TO ix_ctl_datasets_id")

    # Evaluations
    op.rename_table("evaluations", "ctl_evaluations")
    op.execute("ALTER INDEX evaluations_pkey RENAME TO ctl_evaluations_pkey")
    op.execute(
        "ALTER INDEX ix_evaluations_fides_key RENAME TO ix_ctl_evaluations_fides_key"
    )
    op.execute("ALTER INDEX ix_evaluations_id RENAME TO ix_ctl_evaluations_id")

    # Organizations
    op.rename_table("organizations", "ctl_organizations")
    op.execute("ALTER INDEX organizations_pkey RENAME TO ctl_organizations_pkey")
    op.execute(
        "ALTER INDEX ix_organizations_fides_key RENAME TO ix_ctl_organizations_fides_key"
    )
    op.execute("ALTER INDEX ix_organizations_id RENAME TO ix_ctl_organizations_id")

    # Policies
    op.rename_table("policies", "ctl_policies")
    op.execute("ALTER INDEX policies_pkey RENAME TO ctl_policies_pkey")
    op.execute("ALTER INDEX ix_policies_fides_key RENAME TO ix_ctl_policies_fides_key")
    op.execute("ALTER INDEX ix_policies_id RENAME TO ix_ctl_policies_id")

    # Registries
    op.rename_table("registries", "ctl_registries")
    op.execute("ALTER INDEX registries_pkey RENAME TO ctl_registries_pkey")
    op.execute(
        "ALTER INDEX ix_registries_fides_key RENAME TO ix_ctl_registries_fides_key"
    )
    op.execute("ALTER INDEX ix_registries_id RENAME TO ix_ctl_registries_id")

    # Systems
    op.rename_table("systems", "ctl_systems")
    op.execute("ALTER INDEX systems_pkey RENAME TO ctl_systems_pkey")
    op.execute("ALTER INDEX ix_systems_fides_key RENAME TO ix_ctl_systems_fides_key")
    op.execute("ALTER INDEX ix_systems_id RENAME TO ix_ctl_systems_id")


def downgrade():

    op.rename_table("ctl_data_categories", "data_categories")
    op.execute("ALTER INDEX ctl_data_categories_pkey RENAME TO data_categories_pkey")
    op.execute(
        "ALTER INDEX ix_ctl_data_categories_fides_key RENAME TO ix_data_categories_fides_key"
    )
    op.execute("ALTER INDEX ix_ctl_data_categories_id RENAME TO ix_data_categories_id")

    # Data Subjects
    op.rename_table("ctl_data_subjects", "data_subjects")
    op.execute("ALTER INDEX ctl_data_subjects_pkey RENAME TO data_subjects_pkey")
    op.execute(
        "ALTER INDEX ix_ctl_data_subjects_fides_key RENAME TO ix_data_subjects_fides_key"
    )
    op.execute("ALTER INDEX ix_ctl_data_subjects_id RENAME TO ix_data_subjects_id")

    # Data Uses
    op.rename_table("ctl_data_uses", "data_uses")
    op.execute("ALTER INDEX ctl_data_uses_pkey RENAME TO data_uses_pkey")
    op.execute(
        "ALTER INDEX ix_ctl_data_uses_fides_key RENAME TO ix_data_uses_fides_key"
    )
    op.execute("ALTER INDEX ix_ctl_data_uses_id RENAME TO ix_data_uses_id")

    # Data Qualifiers
    op.rename_table("ctl_data_qualifiers", "data_qualifiers")
    op.execute("ALTER INDEX ctl_data_qualifiers_pkey RENAME TO data_qualifiers_pkey")
    op.execute(
        "ALTER INDEX ix_ctl_data_qualifiers_fides_key RENAME TO ix_data_qualifiers_fides_key"
    )
    op.execute("ALTER INDEX ix_ctl_data_qualifiers_id RENAME TO ix_data_qualifiers_id")

    # Datasets
    op.rename_table("ctl_datasets", "datasets")
    op.execute("ALTER INDEX ctl_datasets_pkey RENAME TO datasets_pkey")
    op.execute("ALTER INDEX ix_ctl_datasets_fides_key RENAME TO ix_datasets_fides_key")
    op.execute("ALTER INDEX ix_ctl_datasets_id RENAME TO ix_datasets_id")

    # Evaluations
    op.rename_table("ctl_evaluations", "evaluations")
    op.execute("ALTER INDEX ctl_evaluations_pkey RENAME TO evaluations_pkey")
    op.execute(
        "ALTER INDEX ix_ctl_evaluations_fides_key RENAME TO ix_evaluations_fides_key"
    )
    op.execute("ALTER INDEX ix_ctl_evaluations_id RENAME TO ix_evaluations_id")

    # Organizations
    op.rename_table("ctl_organizations", "organizations")
    op.execute("ALTER INDEX ctl_organizations_pkey RENAME TO organizations_pkey")
    op.execute(
        "ALTER INDEX ix_ctl_organizations_fides_key RENAME TO ix_organizations_fides_key"
    )
    op.execute("ALTER INDEX ix_ctl_organizations_id RENAME TO ix_organizations_id")

    # Policies
    op.rename_table("ctl_policies", "policies")
    op.execute("ALTER INDEX ctl_policies_pkey RENAME TO policies_pkey")
    op.execute("ALTER INDEX ix_ctl_policies_fides_key RENAME TO ix_policies_fides_key")
    op.execute("ALTER INDEX ix_ctl_policies_id RENAME TO ix_policies_id")

    # Registries
    op.rename_table("ctl_registries", "registries")
    op.execute("ALTER INDEX ctl_registries_pkey RENAME TO registries_pkey")
    op.execute(
        "ALTER INDEX ix_ctl_registries_fides_key RENAME TO ix_registries_fides_key"
    )
    op.execute("ALTER INDEX ix_ctl_registries_id RENAME TO ix_registries_id")

    # Systems
    op.rename_table("ctl_systems", "systems")
    op.execute("ALTER INDEX ctl_systems_pkey RENAME TO systems_pkey")
    op.execute("ALTER INDEX ix_ctl_systems_fides_key RENAME TO ix_systems_fides_key")
    op.execute("ALTER INDEX ix_ctl_systems_id RENAME TO ix_systems_id")



# type: ignore
# pylint: disable=redefined-outer-name,cell-var-from-loop

"""
Contains all of the generic CRUD endpoints that can be
generated programmatically for each resource.
"""

from typing import Dict, List

from fastapi import Depends, HTTPException, Response, Security, status
from fideslang import Dataset, FidesModel, model_map
from fideslang.validation import FidesKey
from pydantic import ValidationError as PydanticValidationError
from sqlalchemy.ext.asyncio import AsyncSession
from starlette.status import HTTP_422_UNPROCESSABLE_ENTITY

from fides.api.ctl.database.crud import (
    create_resource,
    delete_resource,
    get_resource,
    list_resource,
    update_resource,
    upsert_resources,
)
from fides.api.ctl.database.session import get_async_db
from fides.api.ctl.routes.util import (
    API_PREFIX,
    forbid_if_default,
    forbid_if_editing_any_is_default,
    forbid_if_editing_is_default,
    get_resource_type,
)
from fides.api.ctl.sql_models import (
    DataCategory,
    models_with_default_field,
    sql_model_map,
)
from fides.api.ctl.utils import errors
from fides.api.ctl.utils.api_router import APIRouter
from fides.api.ops.api.v1 import scope_registry
from fides.api.ops.schemas.dataset import validate_data_categories_against_db
from fides.api.ops.util.oauth_util import verify_oauth_client_cli


async def get_data_categories_from_db(async_session: AsyncSession) -> List[FidesKey]:
    """Similar method to one on the ops side except this uses an async session to retrieve data categories"""
    resources = await list_resource(DataCategory, async_session)
    data_categories = [res.fides_key for res in resources]
    return data_categories


async def validate_data_categories(
    resource: FidesModel, async_session: AsyncSession
) -> None:
    """Validate data categories defined on Datasets against data categories in the db"""
    if not isinstance(resource, Dataset):
        return

    try:
        defined_data_categories: List[FidesKey] = await get_data_categories_from_db(
            async_session
        )
        validate_data_categories_against_db(resource, defined_data_categories)
    except PydanticValidationError as e:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY, detail=e.errors()
        )


# CRUD Endpoints
routers: List[APIRouter] = []
for model_type, fides_model in model_map.items():
    # Programmatically define routers for each resource type
    router = APIRouter(
        tags=[fides_model.__name__],
        prefix=f"{API_PREFIX}/{model_type}",
    )

    @router.post(
        "/",
        response_model=fides_model,
        status_code=status.HTTP_201_CREATED,
        dependencies=[
            Security(
                verify_oauth_client_cli, scopes=[scope_registry.CLI_OBJECTS_CREATE]
            )
        ],
        responses={
            status.HTTP_403_FORBIDDEN: {
                "content": {
                    "application/json": {
                        "example": {
                            "detail": {
                                "error": "user does not have permission to modify this resource",
                                "resource_type": model_type,
                                "fides_key": "example.key",
                            }
                        }
                    }
                }
            },
        },
    )
    async def create(
        resource: fides_model,
        resource_type: str = get_resource_type(router),
        db: AsyncSession = Depends(get_async_db),
    ) -> Dict:
        """
        Create a resource.

        Payloads with an is_default field can only be set to False,
        will return a `403 Forbidden`.
        """
        sql_model = sql_model_map[resource_type]
        await validate_data_categories(resource, db)
        if sql_model in models_with_default_field and resource.is_default:
            raise errors.ForbiddenError(resource_type, resource.fides_key)
        return await create_resource(sql_model, resource.dict(), db)

    @router.get(
        "/",
        dependencies=[
            Security(verify_oauth_client_cli, scopes=[scope_registry.CLI_OBJECTS_READ])
        ],
        response_model=List[fides_model],
        name="List",
    )
    async def ls(  # pylint: disable=invalid-name
        resource_type: str = get_resource_type(router),
        db: AsyncSession = Depends(get_async_db),
    ) -> List:
        """Get a list of all of the resources of this type."""
        sql_model = sql_model_map[resource_type]
        return await list_resource(sql_model, db)

    @router.get(
        "/{fides_key}",
        dependencies=[
            Security(verify_oauth_client_cli, scopes=[scope_registry.CLI_OBJECTS_READ])
        ],
        response_model=fides_model,
    )
    async def get(
        fides_key: str,
        resource_type: str = get_resource_type(router),
        db: AsyncSession = Depends(get_async_db),
    ) -> Dict:
        """Get a resource by its fides_key."""
        sql_model = sql_model_map[resource_type]
        return await get_resource(sql_model, fides_key, db)

    @router.put(
        "/",
        response_model=fides_model,
        dependencies=[
            Security(
                verify_oauth_client_cli, scopes=[scope_registry.CLI_OBJECTS_UPDATE]
            )
        ],
        responses={
            status.HTTP_403_FORBIDDEN: {
                "content": {
                    "application/json": {
                        "example": {
                            "detail": {
                                "error": "user does not have permission to modify this resource",
                                "resource_type": model_type,
                                "fides_key": "example.key",
                            }
                        }
                    }
                }
            },
        },
    )
    async def update(
        resource: fides_model,
        resource_type: str = get_resource_type(router),
        db: AsyncSession = Depends(get_async_db),
    ) -> Dict:
        """
        Update a resource by its fides_key.

        The `is_default` field cannot be updated and will respond
        with a `403 Forbidden` if attempted.
        """
        sql_model = sql_model_map[resource_type]
        await validate_data_categories(resource, db)
        await forbid_if_editing_is_default(sql_model, resource.fides_key, resource, db)
        return await update_resource(sql_model, resource.dict(), db)

    @router.post(
        "/upsert",
        dependencies=[
            Security(
                verify_oauth_client_cli,
                scopes=[
                    scope_registry.CLI_OBJECTS_CREATE,
                    scope_registry.CLI_OBJECTS_UPDATE,
                ],
            )
        ],
        responses={
            status.HTTP_200_OK: {
                "content": {
                    "application/json": {
                        "example": {
                            "message": f"Upserted 3 {model_type}(s)",
                            "inserted": 0,
                            "updated": 3,
                        }
                    }
                }
            },
            status.HTTP_201_CREATED: {
                "content": {
                    "application/json": {
                        "example": {
                            "message": f"Upserted 3 {model_type}(s)",
                            "inserted": 1,
                            "updated": 2,
                        }
                    }
                }
            },
            status.HTTP_403_FORBIDDEN: {
                "content": {
                    "application/json": {
                        "example": {
                            "detail": {
                                "error": "user does not have permission to modify this resource",
                                "resource_type": "DataCategory",
                                "fides_key": "example.key",
                            }
                        }
                    }
                }
            },
        },
    )
    async def upsert(
        resources: List[fides_model],
        response: Response,
        resource_type: str = get_resource_type(router),
        db: AsyncSession = Depends(get_async_db),
    ) -> Dict:
        """
        For any resource in `resources` that already exists in the database,
        update the resource by its `fides_key`. Otherwise, create a new resource.

        Responds with a `201 Created` if even a single resource in `resources`
        did not previously exist. Otherwise, responds with a `200 OK`.

        The `is_default` field cannot be updated and will respond
        with a `403 Forbidden` if attempted.
        """

        sql_model = sql_model_map[resource_type]
        resource_dicts = [resource.dict() for resource in resources]
        for resource in resources:
            await validate_data_categories(resource, db)

        await forbid_if_editing_any_is_default(sql_model, resource_dicts, db)
        result = await upsert_resources(sql_model, resource_dicts, db)
        response.status_code = (
            status.HTTP_201_CREATED if result[0] > 0 else response.status_code
        )

        return {
            "message": f"Upserted {len(resources)} {sql_model.__name__}(s)",
            "inserted": result[0],
            "updated": result[1],
        }

    @router.delete(
        "/{fides_key}",
        dependencies=[
            Security(
                verify_oauth_client_cli, scopes=[scope_registry.CLI_OBJECTS_DELETE]
            )
        ],
        responses={
            status.HTTP_403_FORBIDDEN: {
                "content": {
                    "application/json": {
                        "example": {
                            "detail": {
                                "error": "user does not have permission to modify this resource",
                                "resource_type": model_type,
                                "fides_key": "example.key",
                            }
                        }
                    }
                }
            },
        },
    )
    async def delete(
        fides_key: str,
        resource_type: str = get_resource_type(router),
        db: AsyncSession = Depends(get_async_db),
    ) -> Dict:
        """
        Delete a resource by its fides_key.

        Resources that have `is_default=True` cannot be deleted and will respond
        with a `403 Forbidden`.
        """
        sql_model = sql_model_map[resource_type]
        await forbid_if_default(sql_model, fides_key, db)
        deleted_resource = await delete_resource(sql_model, fides_key, db)
        # Convert the resource to a dict explicitly for the response
        deleted_resource_dict = (
            model_map[resource_type].from_orm(deleted_resource).dict()
        )
        return {
            "message": "resource deleted",
            "resource": deleted_resource_dict,
        }

    routers += [router]



"""Remove datasetconfig dataset

Revision ID: d6c6c6555c86
Revises: 9c6f62e4c9da
Create Date: 2022-12-20 20:44:32.840423

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "d6c6c6555c86"
down_revision = "9c6f62e4c9da"
branch_labels = None
depends_on = None


def upgrade():
    op.drop_column("datasetconfig", "dataset")


def downgrade():
    op.add_column(
        "datasetconfig",
        sa.Column(
            "dataset",
            postgresql.JSONB(astext_type=sa.Text()),
            autoincrement=False,
            nullable=False,
        ),
    )



from __future__ import annotations

from typing import Any, Callable, Dict, List, Set, Tuple, cast

import pydash.collections
from loguru import logger

from fides.api.ops.common_exceptions import TraversalError
from fides.api.ops.graph.config import (
    ROOT_COLLECTION_ADDRESS,
    Collection,
    CollectionAddress,
    Field,
    FieldAddress,
    FieldPath,
    GraphDataset,
)
from fides.api.ops.graph.graph import DatasetGraph, Edge, Node
from fides.api.ops.util.collection_util import Row, append
from fides.api.ops.util.matching_queue import MatchingQueue

Datastore = Dict[CollectionAddress, List[Row]]
"""A type expressing retrieved rows of data from a specified collection"""


class TraversalNode:
    """Base traversal traversal_node type. This type will never be used directly."""

    def __init__(self, node: Node):
        self.node = node
        self.address = node.address
        # address to a child that holds both a traversal_node and all possible paths
        # from this traversal_node to it: {address -> [(traversal_node, from_field, to_field)}
        self.children: Dict[
            CollectionAddress, List[Tuple[TraversalNode, FieldPath, FieldPath]]
        ] = {}
        self.parents: Dict[
            CollectionAddress, List[Tuple[TraversalNode, FieldPath, FieldPath]]
        ] = {}
        self.is_terminal_node = False

    def add_child(self, child_node: TraversalNode, edge: Edge) -> None:
        """Add other as a child to this traversal_node along the provided edge."""
        addresses = edge.split_by_address(self.address)  # (traversal_node -> other)
        if addresses:
            self_field_address, other_field_address = addresses
            append(
                self.children,
                other_field_address.collection_address(),
                (
                    child_node,
                    self_field_address.field_path,
                    other_field_address.field_path,
                ),
            )
            append(
                child_node.parents,
                self_field_address.collection_address(),
                (self, self_field_address.field_path, other_field_address.field_path),
            )

    def incoming_edges(self) -> Set[Edge]:
        """Return the incoming edges to this traversal_node,in (other.address -> self.address) order."""
        return {
            Edge(
                p_collection_address.field_address(parent_field_path),
                self.address.field_address(self_field_path),
            )
            for p_collection_address, tuples in self.parents.items()
            for _, parent_field_path, self_field_path in tuples
        }

    def incoming_edges_from_same_dataset(self) -> Set[Edge]:
        """Return the incoming edges from the same dataset"""
        return {
            Edge(
                p_collection_address.field_address(parent_field_path),
                self.address.field_address(self_field_path),
            )
            for p_collection_address, tuples in self.parents.items()
            if p_collection_address.dataset == self.address.dataset
            for _, parent_field_path, self_field_path in tuples
        }

    def outgoing_edges(self) -> Set[Edge]:
        """Return the outgoing edges to this traversal_node,in (self.address -> other.address) order."""
        return {
            Edge(
                self.address.field_address(self_field_path),
                c_collection_address.field_address(child_field_path),
            )
            for c_collection_address, tuples in self.children.items()
            for _, self_field_path, child_field_path in tuples
        }

    @property
    def query_field_paths(self) -> Set[FieldPath]:
        """
        All of the possible field paths that we can query for possible filter values.
        These are field paths that are the ends of incoming edges.
        """
        return {edge.f2.field_path for edge in self.incoming_edges()}

    def typed_filtered_values(self, input_data: Dict[str, List[Any]]) -> Dict[str, Any]:
        """
        Return a filtered list of key/value sets of data items that are both in
        the list of incoming edge fields, and contain data in the input data set.

        The values are cast based on field types, if those types are specified.
        """
        out = {}
        for key, values in input_data.items():
            path: FieldPath = FieldPath.parse(key)
            field: Field | None = self.node.collection.field(path)

            if field and path in self.query_field_paths and isinstance(values, list):
                cast_values = [field.cast(v) for v in values]
                filtered = list(filter(lambda x: x is not None, cast_values))
                if filtered:
                    out[key] = filtered
        return out

    def can_run_given(self, remaining_node_keys: Set[CollectionAddress]) -> bool:
        """True if finished_node_keys covers all the nodes that this traversal_node is waiting for.  If
        all nodes this traversal_node is waiting for have finished, it's ok for this traversal_node to run.
        """
        if self.node.collection.after.intersection(
            remaining_node_keys
        ) or self.node.dataset.after.intersection(
            {k.dataset for k in remaining_node_keys}
        ):
            return False
        return True

    def is_root_node(self) -> bool:
        """This traversal_node is the defined traversal start"""
        return self.address == ROOT_COLLECTION_ADDRESS

    def debug(self) -> Dict[str, Any]:
        """Generate debug descriptive output of the traversal."""
        _from: Dict[str, Set[str]] = {}

        for edge in self.incoming_edges():
            foreign = edge.f1
            local = edge.f2
            # if we have a bidirectional edge, make sure that we correctly label
            # which edge is foreign.
            if foreign.collection_address() == self.address:
                foreign, local = local, foreign
            key = str(foreign.collection_address())
            val = f"{foreign.field_path.string_path} -> {local.field_path.string_path}"
            if key in _from:
                _from[key].add(val)
            else:
                _from[key] = {val}
        to: Dict[str, Set[str]] = {}
        for k, v in self.children.items():
            to[str(k)] = {f"{f[1].string_path} -> {f[2].string_path}" for f in v}
        return {
            "from": {k: set(v) for k, v in _from.items()},
            "to": {k: set(v) for k, v in to.items()},
        }


def artificial_traversal_node(address: CollectionAddress) -> TraversalNode:
    """generate an 'artificial' traversal_node pointing to the given address. This is used to
    generate artificial root and termination nodes that correspond to just an address, but
    have no actual corresponding collection dataset"""
    ds: Collection = Collection(name=address.collection, fields=[])
    node = Node(
        GraphDataset(
            name=address.dataset, collections=[ds], connection_key="__IGNORE__"
        ),
        ds,
    )
    return TraversalNode(node)


class Traversal:
    """Handling for a single reified traversal of a graph based on input (seed) data."""

    def extract_seed_field_addresses(
        self,
    ) -> Dict[FieldAddress, str]:
        """Find seeded values in the graph based on the input seed and return
        a list of the populated seed nodes"""

        return {
            identity_address: seed_key
            for identity_address, seed_key in self.graph.identity_keys.items()
            if seed_key in self.seed_data
        }

    def __init__(self, graph: DatasetGraph, data: Dict[str, Any]):
        self.graph = graph
        self.seed_data = data
        self.traversal_node_dict = {k: TraversalNode(v) for k, v in graph.nodes.items()}
        self.edges: Set[Edge] = graph.edges.copy()
        self.root_node = artificial_traversal_node(ROOT_COLLECTION_ADDRESS)
        for (
            start_field_address,
            seed_key,
        ) in self.extract_seed_field_addresses().items():
            self.edges.add(
                Edge(
                    FieldAddress(
                        ROOT_COLLECTION_ADDRESS.dataset,
                        ROOT_COLLECTION_ADDRESS.collection,
                        seed_key,
                    ),
                    start_field_address,
                )
            )

        self.__verify_traversal()

    def __verify_traversal(self) -> None:
        """Verify that a valid traversal exists. This method simply assembles a traversal
        and raises an error on any traversal failure conditions."""
        self.traverse(
            {self.root_node.address: [self.seed_data]},
            lambda n, m: logger.info("Traverse {}", n.address),
        )

    def traversal_map(
        self,
    ) -> Tuple[Dict[str, List[Dict[str, Any]]], List[CollectionAddress]]:
        """Generate a descriptive map of the traversal generated.

        For each traversal_node N, list
         { N.address:
             {"from": {traversal_node address:
                       [field in "from", field in N for all edges from "from" to N]}}
             {"to": {traversal_node address,
                       [field in N, field in "to"" for all edges from N to "to"]}}

        """

        def traversal_dict_fn(
            tn: TraversalNode, data: Dict[CollectionAddress, Dict[str, Any]]
        ) -> None:
            data[tn.address] = tn.debug()

        db = {ROOT_COLLECTION_ADDRESS: [self.seed_data]}
        traversal_ends = self.traverse(db, traversal_dict_fn)

        return {str(k): v for k, v in db.items()}, traversal_ends

    def traverse(  # pylint: disable=R0914
        self,
        environment: Dict[CollectionAddress, Any],
        node_run_fn: Callable[[TraversalNode, Dict[CollectionAddress, Any]], None],
    ) -> List[CollectionAddress]:
        """Traverse and call run() on each traversal_node in turn. T represents an environment that
        can provide or collect values as each traversal_node is run.

        Returns a list of termination traversal_node addresses so that we can take action on completed
        traversal.



        We define the root traversal_node as a traversal_node whose children are any nodes that have identity (seed)
        data.
        We start with
        - a queue holding only the root traversal_node.
        - a (copied) set of all of the edges in the graph.

        - Pop the first eligible traversal_node from the queue.
        - call the passed in callable node_run_function on this traversal_node. Mark this traversal_node as "finished"
        - Delete all edges from any finished nodes to this traversal_node.
        - put all of this nodes children in the queue.

        Some nodes have conditions, like "don't run me until after traversal_node X". When we pop a value from
        the queue, we are taking this into account. If the queue contains nodes, but none of them are
        eligible (e.g. Node A can't run until after B, and traversal_node B that can't run until after A)
        raise a TraversalError.

        We also raise a TraversalError if the queue is empty but some nodes have not been visited. In
        that case they are unreachable.

        """
        if environment:
            logger.info(
                "starting traversal",
            )
        remaining_node_keys: Set[CollectionAddress] = set(
            self.traversal_node_dict.keys()
        )
        finished_nodes: dict[CollectionAddress, TraversalNode] = {}
        running_node_queue: MatchingQueue[TraversalNode] = MatchingQueue(self.root_node)
        remaining_edges: Set[Edge] = self.edges.copy()
        while not running_node_queue.is_empty():

            # this is to support the "run traversal_node A AFTER traversal_node B functionality:"
            n = running_node_queue.pop_first_match(
                lambda x: x.can_run_given(remaining_node_keys)
            )

            if n:
                node_run_fn(n, environment)
                # delete all edges between the traversal_node that's just run and any completed nodes
                for finished_node_address, finished_node in finished_nodes.items():
                    completed_edges = Edge.delete_edges(
                        remaining_edges,
                        finished_node_address,
                        cast(TraversalNode, n).address,  # type: ignore[redundant-cast]
                    )
                    # append edges that end in this traversal_node
                    for edge in filter(
                        lambda _edge: _edge.ends_with_collection(
                            cast(TraversalNode, n).address  # type: ignore[redundant-cast]
                        ),
                        completed_edges,
                    ):
                        # note, this will not work for self-reference
                        finished_node.add_child(n, edge)
                # next edges = take all edges including n that are _not_ in edges_from_completed_nodes
                # in the form (field_address_this, field_address_foreign)

                edges_to_children = pydash.collections.filter_(
                    [
                        e.split_by_address(cast(TraversalNode, n).address)  # type: ignore[redundant-cast]
                        for e in remaining_edges
                        if e.contains(n.address)
                    ]
                )
                if not edges_to_children:
                    n.is_terminal_node = True

                # child traversal_node addresses are the address portion of the above
                child_node_addresses = {
                    a[1].collection_address() for a in edges_to_children if a
                }
                for nxt_address in child_node_addresses:
                    # only add the next traversal_node to the queue if it is not already there (no duplicates)
                    running_node_queue.push_if_new(
                        self.traversal_node_dict[nxt_address]
                    )
                finished_nodes[n.address] = n
                remaining_node_keys.difference_update({n.address})
            else:
                # traversal traversal_node dict diff finished nodes
                logger.error(
                    "Node could not be reached given specified ordering [{}]",
                    ",".join([str(tn.address) for tn in running_node_queue.data]),
                )
                raise TraversalError(
                    f"""Node could not be reached given the specified ordering:
                    [{','.join([str(tn.address) for tn in running_node_queue.data])}]"""
                )

        # error if there are nodes that have not been visited
        if remaining_node_keys:
            logger.error(
                "Some nodes were not reachable: {}",
                ",".join([str(x) for x in remaining_node_keys]),
            )
            raise TraversalError(
                f"Some nodes were not reachable: {','.join([str(x) for x in remaining_node_keys])}"
            )
        # error if there are edges that have not been visited
        if remaining_edges:
            logger.error(
                "Some edges were not reachable: {}",
                ",".join([str(x) for x in remaining_edges]),
            )
            raise TraversalError(
                f"Some edges were not reachable: {','.join([str(x) for x in remaining_edges])}"
            )

        end_nodes = [
            tn.address for tn in finished_nodes.values() if tn.is_terminal_node
        ]
        if environment:
            logger.debug("Found {} end nodes: {}", len(end_nodes), end_nodes)
        return end_nodes



"""initial migration

Revision ID: 26934c96ec80
Revises:
Create Date: 2021-09-22 04:12:38.479740

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "26934c96ec80"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "data_categories",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("fides_key", sa.String(), nullable=False),
        sa.Column("organization_fides_key", sa.Text(), nullable=True),
        sa.Column("name", sa.Text(), nullable=True),
        sa.Column("description", sa.Text(), nullable=True),
        sa.Column("parent_key", sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint("id", "fides_key"),
    )
    op.create_index(
        op.f("ix_data_categories_fides_key"),
        "data_categories",
        ["fides_key"],
        unique=True,
    )
    op.create_index(
        op.f("ix_data_categories_id"), "data_categories", ["id"], unique=True
    )
    op.create_table(
        "data_qualifiers",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("fides_key", sa.String(), nullable=False),
        sa.Column("organization_fides_key", sa.Text(), nullable=True),
        sa.Column("name", sa.Text(), nullable=True),
        sa.Column("description", sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint("id", "fides_key"),
    )
    op.create_index(
        op.f("ix_data_qualifiers_fides_key"),
        "data_qualifiers",
        ["fides_key"],
        unique=True,
    )
    op.create_index(
        op.f("ix_data_qualifiers_id"), "data_qualifiers", ["id"], unique=True
    )
    op.create_table(
        "data_subjects",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("fides_key", sa.String(), nullable=False),
        sa.Column("organization_fides_key", sa.Text(), nullable=True),
        sa.Column("name", sa.Text(), nullable=True),
        sa.Column("description", sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint("id", "fides_key"),
    )
    op.create_index(
        op.f("ix_data_subjects_fides_key"), "data_subjects", ["fides_key"], unique=True
    )
    op.create_index(op.f("ix_data_subjects_id"), "data_subjects", ["id"], unique=True)
    op.create_table(
        "data_uses",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("fides_key", sa.String(), nullable=False),
        sa.Column("organization_fides_key", sa.Text(), nullable=True),
        sa.Column("name", sa.Text(), nullable=True),
        sa.Column("description", sa.Text(), nullable=True),
        sa.Column("parent_key", sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint("id", "fides_key"),
    )
    op.create_index(
        op.f("ix_data_uses_fides_key"), "data_uses", ["fides_key"], unique=True
    )
    op.create_index(op.f("ix_data_uses_id"), "data_uses", ["id"], unique=True)
    op.create_table(
        "datasets",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("fides_key", sa.String(), nullable=False),
        sa.Column("organization_fides_key", sa.Text(), nullable=True),
        sa.Column("name", sa.Text(), nullable=True),
        sa.Column("description", sa.Text(), nullable=True),
        sa.Column("meta", sa.JSON(), nullable=True),
        sa.Column("data_categories", sa.ARRAY(sa.String()), nullable=True),
        sa.Column("data_qualifier", sa.ARRAY(sa.String()), nullable=True),
        sa.Column("location", sa.String(), nullable=True),
        sa.Column("dataset_type", sa.String(), nullable=True),
        sa.Column("fields", sa.JSON(), nullable=True),
        sa.PrimaryKeyConstraint("id", "fides_key"),
    )
    op.create_index(
        op.f("ix_datasets_fides_key"), "datasets", ["fides_key"], unique=True
    )
    op.create_index(op.f("ix_datasets_id"), "datasets", ["id"], unique=True)
    op.create_table(
        "evaluations",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("fides_key", sa.String(), nullable=False),
        sa.Column("status", sa.String(), nullable=True),
        sa.Column("details", sa.ARRAY(sa.String()), nullable=True),
        sa.Column("message", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id", "fides_key"),
    )
    op.create_index(
        op.f("ix_evaluations_fides_key"), "evaluations", ["fides_key"], unique=True
    )
    op.create_index(op.f("ix_evaluations_id"), "evaluations", ["id"], unique=True)
    op.create_table(
        "organizations",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("fides_key", sa.String(), nullable=False),
        sa.Column("organization_fides_key", sa.Text(), nullable=True),
        sa.Column("name", sa.Text(), nullable=True),
        sa.Column("description", sa.Text(), nullable=True),
        sa.Column("organization_parent_key", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id", "fides_key"),
    )
    op.create_index(
        op.f("ix_organizations_fides_key"), "organizations", ["fides_key"], unique=True
    )
    op.create_index(op.f("ix_organizations_id"), "organizations", ["id"], unique=True)
    op.create_table(
        "policies",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("fides_key", sa.String(), nullable=False),
        sa.Column("organization_fides_key", sa.Text(), nullable=True),
        sa.Column("name", sa.Text(), nullable=True),
        sa.Column("description", sa.Text(), nullable=True),
        sa.Column("rules", sa.JSON(), nullable=True),
        sa.PrimaryKeyConstraint("id", "fides_key"),
    )
    op.create_index(
        op.f("ix_policies_fides_key"), "policies", ["fides_key"], unique=True
    )
    op.create_index(op.f("ix_policies_id"), "policies", ["id"], unique=True)
    op.create_table(
        "registries",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("fides_key", sa.String(), nullable=False),
        sa.Column("organization_fides_key", sa.Text(), nullable=True),
        sa.Column("name", sa.Text(), nullable=True),
        sa.Column("description", sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint("id", "fides_key"),
    )
    op.create_index(
        op.f("ix_registries_fides_key"), "registries", ["fides_key"], unique=True
    )
    op.create_index(op.f("ix_registries_id"), "registries", ["id"], unique=True)
    op.create_table(
        "systems",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("fides_key", sa.String(), nullable=False),
        sa.Column("organization_fides_key", sa.Text(), nullable=True),
        sa.Column("name", sa.Text(), nullable=True),
        sa.Column("description", sa.Text(), nullable=True),
        sa.Column("registry_id", sa.String(), nullable=True),
        sa.Column("meta", sa.JSON(), nullable=True),
        sa.Column("system_type", sa.String(), nullable=True),
        sa.Column("system_dependencies", sa.ARRAY(sa.String()), nullable=True),
        sa.Column("privacy_declarations", sa.JSON(), nullable=True),
        sa.PrimaryKeyConstraint("id", "fides_key"),
    )
    op.create_index(op.f("ix_systems_fides_key"), "systems", ["fides_key"], unique=True)
    op.create_index(op.f("ix_systems_id"), "systems", ["id"], unique=True)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_systems_id"), table_name="systems")
    op.drop_index(op.f("ix_systems_fides_key"), table_name="systems")
    op.drop_table("systems")
    op.drop_index(op.f("ix_registries_id"), table_name="registries")
    op.drop_index(op.f("ix_registries_fides_key"), table_name="registries")
    op.drop_table("registries")
    op.drop_index(op.f("ix_policies_id"), table_name="policies")
    op.drop_index(op.f("ix_policies_fides_key"), table_name="policies")
    op.drop_table("policies")
    op.drop_index(op.f("ix_organizations_id"), table_name="organizations")
    op.drop_index(op.f("ix_organizations_fides_key"), table_name="organizations")
    op.drop_table("organizations")
    op.drop_index(op.f("ix_evaluations_id"), table_name="evaluations")
    op.drop_index(op.f("ix_evaluations_fides_key"), table_name="evaluations")
    op.drop_table("evaluations")
    op.drop_index(op.f("ix_datasets_id"), table_name="datasets")
    op.drop_index(op.f("ix_datasets_fides_key"), table_name="datasets")
    op.drop_table("datasets")
    op.drop_index(op.f("ix_data_uses_id"), table_name="data_uses")
    op.drop_index(op.f("ix_data_uses_fides_key"), table_name="data_uses")
    op.drop_table("data_uses")
    op.drop_index(op.f("ix_data_subjects_id"), table_name="data_subjects")
    op.drop_index(op.f("ix_data_subjects_fides_key"), table_name="data_subjects")
    op.drop_table("data_subjects")
    op.drop_index(op.f("ix_data_qualifiers_id"), table_name="data_qualifiers")
    op.drop_index(op.f("ix_data_qualifiers_fides_key"), table_name="data_qualifiers")
    op.drop_table("data_qualifiers")
    op.drop_index(op.f("ix_data_categories_id"), table_name="data_categories")
    op.drop_index(op.f("ix_data_categories_fides_key"), table_name="data_categories")
    op.drop_table("data_categories")
    # ### end Alembic commands ###



import copy
import traceback
from abc import ABC
from functools import wraps
from time import sleep
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

import dask
from dask import delayed  # type: ignore[attr-defined]
from dask.threaded import get
from loguru import logger
from sqlalchemy.orm import Session

from fides.api.ops.common_exceptions import (
    CollectionDisabled,
    NotSupportedForCollection,
    PrivacyRequestErasureEmailSendRequired,
    PrivacyRequestPaused,
)
from fides.api.ops.graph.analytics_events import (
    fideslog_graph_rerun,
    prepare_rerun_graph_analytics_event,
)
from fides.api.ops.graph.config import (
    ROOT_COLLECTION_ADDRESS,
    TERMINATOR_ADDRESS,
    CollectionAddress,
    Field,
    FieldAddress,
    FieldPath,
)
from fides.api.ops.graph.graph import DatasetGraph, Edge, Node
from fides.api.ops.graph.graph_differences import format_graph_for_caching
from fides.api.ops.graph.traversal import Traversal, TraversalNode
from fides.api.ops.models.connectionconfig import AccessLevel, ConnectionConfig
from fides.api.ops.models.policy import ActionType, Policy
from fides.api.ops.models.privacy_request import (
    Consent,
    ExecutionLogStatus,
    PrivacyRequest,
)
from fides.api.ops.service.connectors.base_connector import BaseConnector
from fides.api.ops.task.consolidate_query_matches import consolidate_query_matches
from fides.api.ops.task.filter_element_match import filter_element_match
from fides.api.ops.task.refine_target_path import FieldPathNodeInput
from fides.api.ops.task.task_resources import TaskResources
from fides.api.ops.util.cache import get_cache
from fides.api.ops.util.collection_util import NodeInput, Row, append, partition
from fides.api.ops.util.logger import Pii
from fides.api.ops.util.saas_util import FIDESOPS_GROUPED_INPUTS
from fides.core.config import get_config

dask.config.set(scheduler="threads")

COLLECTION_FIELD_PATH_MAP = Dict[CollectionAddress, List[Tuple[FieldPath, FieldPath]]]
CONFIG = get_config()
EMPTY_REQUEST = PrivacyRequest()


def retry(
    action_type: ActionType,
    default_return: Any,
) -> Callable:
    """
    Retry decorator for access and right to forget requests requests -

    If an exception is raised, we retry the function `count` times with exponential backoff. After the number of
    retries have expired, we call GraphTask.end() with the appropriate `action_type` and `default_return`.

    If we exceed the number of TASK_RETRY_COUNT retries, we re-raise the exception to stop execution of the privacy request.
    """

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def result(*args: Any, **kwargs: Any) -> Any:
            func_delay = CONFIG.execution.task_retry_delay
            method_name = func.__name__
            self = args[0]

            raised_ex: Optional[Union[BaseException, Exception]] = None
            for attempt in range(CONFIG.execution.task_retry_count + 1):
                try:
                    self.skip_if_disabled()
                    # Create ExecutionLog with status in_processing or retrying
                    if attempt:
                        self.log_retry(action_type)
                    else:
                        self.log_start(action_type)
                    # Run access or erasure request
                    return func(*args, **kwargs)
                except PrivacyRequestPaused as ex:
                    logger.warning(
                        "Privacy request {} paused {}",
                        method_name,
                        self.traversal_node.address,
                    )
                    self.log_paused(action_type, ex)
                    # Re-raise to stop privacy request execution on pause.
                    raise
                except PrivacyRequestErasureEmailSendRequired as exc:
                    self.log_end(action_type, ex=None, success_override_msg=exc)
                    self.resources.cache_erasure(
                        f"{self.traversal_node.address.value}", 0
                    )  # Cache that the erasure was performed in case we need to restart
                    return 0
                except (CollectionDisabled, NotSupportedForCollection) as exc:
                    logger.warning(
                        "Skipping collection {} for privacy_request: {}",
                        self.traversal_node.address,
                        self.resources.request.id,
                    )
                    self.log_skipped(action_type, exc)
                    return default_return
                except BaseException as ex:  # pylint: disable=W0703
                    func_delay *= CONFIG.execution.task_retry_backoff
                    logger.warning(
                        "Retrying {} {} in {} seconds...",
                        method_name,
                        self.traversal_node.address,
                        func_delay,
                    )
                    sleep(func_delay)
                    raised_ex = ex

            self.log_end(action_type, raised_ex)
            self.resources.request.cache_failed_checkpoint_details(
                step=action_type, collection=self.traversal_node.address
            )
            # Re-raise to stop privacy request execution on failure.
            raise raised_ex  # type: ignore

        return result

    return decorator


class GraphTask(ABC):  # pylint: disable=too-many-instance-attributes
    """A task that operates on one traversal_node of a traversal"""

    def __init__(
        self, traversal_node: TraversalNode, resources: TaskResources
    ):  # cache config, log config, db store config
        super().__init__()
        self.traversal_node = traversal_node
        self.resources = resources
        self.connector: BaseConnector = resources.get_connector(
            self.traversal_node.node.dataset.connection_key  # ConnectionConfig.key
        )

        # build incoming edges to the form : [dataset address: [(foreign field, local field)]
        self.incoming_edges_by_collection: Dict[
            CollectionAddress, List[Edge]
        ] = partition(
            self.traversal_node.incoming_edges(), lambda e: e.f1.collection_address()
        )

        # the input keys this task will read from.These will build the dask graph
        self.input_keys: List[CollectionAddress] = sorted(
            self.incoming_edges_by_collection.keys()
        )

        self.key = self.traversal_node.address

        self.execution_log_id = None
        # a local copy of the execution log record written to. If we write multiple status
        # updates, we will use this id to ensure that we're updating rather than creating
        # a new record

    def __repr__(self) -> str:
        return f"{type(self)}:{self.key}"

    @property
    def grouped_fields(self) -> Set[str]:
        """Convenience property - returns a set of fields that have been specified on the collection as dependent
        upon one another
        """
        return self.traversal_node.node.collection.grouped_inputs or set()

    @property
    def dependent_identity_fields(self) -> bool:
        """If the current collection needs inputs from other collections, in addition to its seed data."""
        collection = self.traversal_node.node.collection
        for field in self.grouped_fields:
            if collection.field(FieldPath(field)).identity:  # type: ignore
                return True
        return False

    def build_incoming_field_path_maps(
        self, group_dependent_fields: bool = False
    ) -> Tuple[COLLECTION_FIELD_PATH_MAP, COLLECTION_FIELD_PATH_MAP]:
        """
        For each collection connected to the current collection, return a list of tuples
        mapping the foreign field to the local field.  This is used to process data from incoming collections
        into the current collection.

        :param group_dependent_fields: Whether we should split the incoming fields into two groups: one whose
        fields are completely independent of one another, and the other whose incoming data needs to stay linked together.
        If False, all fields are returned in the first tuple, and the second tuple just maps collections to an empty list.

        """

        def field_map(keep: Callable) -> COLLECTION_FIELD_PATH_MAP:
            return {
                col_addr: [
                    (edge.f1.field_path, edge.f2.field_path)
                    for edge in edge_list
                    if keep(edge.f2.field_path.string_path)
                ]
                for col_addr, edge_list in self.incoming_edges_by_collection.items()
            }

        if group_dependent_fields:
            return field_map(
                lambda string_path: string_path not in self.grouped_fields
            ), field_map(lambda string_path: string_path in self.grouped_fields)

        return field_map(lambda string_path: True), field_map(lambda string_path: False)

    def generate_dry_run_query(self) -> Optional[str]:
        """Type-specific query generated for this traversal_node."""
        return self.connector.dry_run_query(self.traversal_node)

    def can_write_data(self) -> bool:
        """Checks if the relevant ConnectionConfig has been granted "write" access to its data"""
        connection_config: ConnectionConfig = self.connector.configuration
        return connection_config.access == AccessLevel.write

    def _combine_seed_data(
        self,
        *data: List[Row],
        grouped_data: Dict[str, Any],
        dependent_field_mappings: COLLECTION_FIELD_PATH_MAP,
    ) -> Dict[str, Any]:
        """Combine the seed data with the other dependent inputs. This is used when the seed data in a collection requires
        inputs from another collection to generate subsequent queries."""
        # Get the identity values from the seeds that were passed into this collection.
        seed_index = self.input_keys.index(ROOT_COLLECTION_ADDRESS)
        seed_data = data[seed_index]

        for (foreign_field_path, local_field_path) in dependent_field_mappings[
            ROOT_COLLECTION_ADDRESS
        ]:
            dependent_values = consolidate_query_matches(
                row=seed_data, target_path=foreign_field_path  # type: ignore
            )
            grouped_data[local_field_path.string_path] = dependent_values
        return grouped_data

    def pre_process_input_data(
        self, *data: List[Row], group_dependent_fields: bool = False
    ) -> NodeInput:
        """
        Consolidates the outputs of queries from potentially multiple collections whose
        data is needed as input into the current collection.

        Each dict in the input list represents the output of a dependent task.
        These outputs should correspond to the input key order.  Any nested fields are
        converted into dot-separated paths in the return.

         table1: [{x:1, y:A}, {x:2, y:B}], table2: [{x:3},{x:4}], table3: [{z: {a: C}, "y": [4, 5]}]
           where table1.x => self.id,
           table1.y=> self.name,
           table2.x=>self.id
           table3.z.a => self.contact.address
           table3.y => self.contact.email
         becomes
         {id:[1,2,3,4], name:["A","B"], contact.address:["C"], "contact.email": [4, 5]}

         If there are dependent fields from one collection into another, they are separated out as follows:
         {fidesops_grouped_inputs: [{"organization_id": 1, "project_id": "math}, {"organization_id": 5, "project_id": "science"}]
        """
        if not len(data) == len(self.input_keys):
            logger.warning(
                "{} expected {} input keys, received {}",
                self,
                len(self.input_keys),
                len(data),
            )

        output: Dict[str, List[Any]] = {FIDESOPS_GROUPED_INPUTS: []}

        (
            independent_field_mappings,
            dependent_field_mappings,
        ) = self.build_incoming_field_path_maps(group_dependent_fields)

        for i, rowset in enumerate(data):
            collection_address = self.input_keys[i]

            if (
                group_dependent_fields
                and self.dependent_identity_fields
                and collection_address == ROOT_COLLECTION_ADDRESS
            ):
                # Skip building data for the root collection if the seed data needs to be combined with other inputs
                continue

            logger.info(
                "Consolidating incoming data into {} from {}.",
                self.traversal_node.node.address,
                collection_address,
            )
            for row in rowset:
                # Consolidate lists of independent field inputs
                for foreign_field_path, local_field_path in independent_field_mappings[
                    collection_address
                ]:
                    new_values: List = consolidate_query_matches(
                        row=row, target_path=foreign_field_path
                    )
                    if new_values:
                        append(output, local_field_path.string_path, new_values)

                # Separately group together dependent inputs if applicable
                if dependent_field_mappings[collection_address]:
                    grouped_data: Dict[str, Any] = {}
                    for (
                        foreign_field_path,
                        local_field_path,
                    ) in dependent_field_mappings[collection_address]:
                        dependent_values: List = consolidate_query_matches(
                            row=row, target_path=foreign_field_path
                        )
                        grouped_data[local_field_path.string_path] = dependent_values

                    if self.dependent_identity_fields:
                        grouped_data = self._combine_seed_data(
                            *data,
                            grouped_data=grouped_data,
                            dependent_field_mappings=dependent_field_mappings,
                        )

                    output[FIDESOPS_GROUPED_INPUTS].append(grouped_data)
        return output

    def update_status(
        self,
        msg: str,
        fields_affected: Any,
        action_type: ActionType,
        status: ExecutionLogStatus,
    ) -> None:
        """Update status activities"""
        self.resources.write_execution_log(
            self.traversal_node.node.dataset.connection_key,
            self.traversal_node.address,
            fields_affected,
            action_type,
            status,
            msg,
        )

    def log_start(self, action_type: ActionType) -> None:
        """Task start activities"""
        logger.info(
            "Starting {}, traversal_node {}", self.resources.request.id, self.key
        )

        self.update_status(
            "starting", [], action_type, ExecutionLogStatus.in_processing
        )

    def log_retry(self, action_type: ActionType) -> None:
        """Task retry activities"""
        logger.info("Retrying {}, node {}", self.resources.request.id, self.key)

        self.update_status("retrying", [], action_type, ExecutionLogStatus.retrying)

    def log_paused(self, action_type: ActionType, ex: Optional[BaseException]) -> None:
        """On paused activities"""
        logger.info("Pausing {}, node {}", self.resources.request.id, self.key)

        self.update_status(str(ex), [], action_type, ExecutionLogStatus.paused)

    def log_skipped(self, action_type: ActionType, ex: str) -> None:
        """Log that a collection was skipped.  For now, this is because a collection has been disabled."""
        logger.info("Skipping {}, node {}", self.resources.request.id, self.key)

        self.update_status(str(ex), [], action_type, ExecutionLogStatus.skipped)

    def log_end(
        self,
        action_type: ActionType,
        ex: Optional[BaseException] = None,
        success_override_msg: Optional[BaseException] = None,
    ) -> None:
        """On completion activities"""
        if ex:
            traceback.print_exc()
            logger.warning(
                "Ending {}, {} with failure {}",
                self.resources.request.id,
                self.key,
                Pii(ex),
            )
            self.update_status(str(ex), [], action_type, ExecutionLogStatus.error)
        else:
            logger.info("Ending {}, {}", self.resources.request.id, self.key)
            self.update_status(
                str(success_override_msg) if success_override_msg else "success",
                build_affected_field_logs(
                    self.traversal_node.node, self.resources.policy, action_type
                ),
                action_type,
                ExecutionLogStatus.complete,
            )

    def post_process_input_data(
        self, pre_processed_inputs: NodeInput
    ) -> FieldPathNodeInput:
        """
        For each entrypoint field, specify if we should return all data, or just data that matches the coerced
        input values. Used for post-processing access request results for a given collection.

        :param pre_processed_inputs: string paths mapped to values that were used to query the current collection
        :return: FieldPaths mapped to type-coerced values that we need to match in
        access request results, or FieldPaths mapped to None if we want to return everything.

        :Example:
        owner.phone field will not be filtered but we will process the owner.identifier results to return
        values that match one of [1234, 5678, 9102]

        {FieldPath("owner", "phone"): None, FieldPath("owner", "identifier"): [1234, 5678, 9102]}
        """
        out: FieldPathNodeInput = {}
        for key, values in pre_processed_inputs.items():
            path: FieldPath = FieldPath.parse(key)
            field: Optional[Field] = self.traversal_node.node.collection.field(path)
            if (
                field
                and path in self.traversal_node.query_field_paths
                and isinstance(values, list)
            ):
                if field.return_all_elements:
                    # All data will be returned
                    out[path] = None
                else:
                    # Default behavior - we will filter values to match those in filtered
                    cast_values = [
                        field.cast(v) for v in values
                    ]  # Cast values to expected type where possible
                    filtered = list(filter(lambda x: x is not None, cast_values))
                    if filtered:
                        out[path] = filtered
        return out

    def access_results_post_processing(
        self, formatted_input_data: NodeInput, output: List[Row]
    ) -> List[Row]:
        """
        Completes post-processing filtering of access request results.

        By default, if an array field was an entry point into the node, return only array elements that *match* the
        condition.  Specifying return_all_elements = true on the field's config will instead return *all* array elements.

        Caches the data in TWO separate formats: 1) erasure format, *replaces* unmatched array elements with placeholder
        text, and 2) access request format, which *removes* unmatched array elements altogether.  If no data was filtered
        out, both cached versions will be the same.
        """
        post_processed_node_input_data: FieldPathNodeInput = (
            self.post_process_input_data(formatted_input_data)
        )

        # For erasures: cache results with non-matching array elements *replaced* with placeholder text
        placeholder_output: List[Row] = copy.deepcopy(output)
        for row in placeholder_output:
            filter_element_match(
                row, query_paths=post_processed_node_input_data, delete_elements=False
            )
        self.resources.cache_results_with_placeholders(
            f"access_request__{self.key}", placeholder_output
        )

        # For access request results, cache results with non-matching array elements *removed*
        for row in output:
            logger.info(
                "Filtering row in {} for matching array elements.",
                self.traversal_node.node.address,
            )
            filter_element_match(row, post_processed_node_input_data)
        self.resources.cache_object(f"access_request__{self.key}", output)

        # Return filtered rows with non-matched array data removed.
        return output

    def skip_if_disabled(self) -> None:
        """Skip execution for the given collection if it is attached to a disabled ConnectionConfig."""
        connection_config: ConnectionConfig = self.connector.configuration
        if connection_config.disabled:
            raise CollectionDisabled(
                f"Skipping collection {self.traversal_node.node.address}. "
                f"ConnectionConfig {connection_config.key} is disabled.",
            )

    @retry(action_type=ActionType.access, default_return=[])
    def access_request(self, *inputs: List[Row]) -> List[Row]:
        """Run an access request on a single node."""
        formatted_input_data: NodeInput = self.pre_process_input_data(
            *inputs, group_dependent_fields=True
        )
        output: List[Row] = self.connector.retrieve_data(
            self.traversal_node,
            self.resources.policy,
            self.resources.request,
            formatted_input_data,
        )
        filtered_output: List[Row] = self.access_results_post_processing(
            self.pre_process_input_data(*inputs, group_dependent_fields=False), output
        )
        self.log_end(ActionType.access)
        return filtered_output

    @retry(action_type=ActionType.erasure, default_return=0)
    def erasure_request(self, retrieved_data: List[Row], *inputs: List[Row]) -> int:
        """Run erasure request"""
        # if there is no primary key specified in the graph node configuration
        # note this in the execution log and perform no erasures on this node
        if not self.traversal_node.node.contains_field(lambda f: f.primary_key):
            logger.warning(
                "No erasures on {} as there is no primary_key defined.",
                self.traversal_node.node.address,
            )
            self.update_status(
                "No values were erased since no primary key was defined for this collection",
                None,
                ActionType.erasure,
                ExecutionLogStatus.complete,
            )
            return 0

        if not self.can_write_data():
            logger.warning(
                "No erasures on {} as its ConnectionConfig does not have write access.",
                self.traversal_node.node.address,
            )
            self.update_status(
                f"No values were erased since this connection {self.connector.configuration.key} has not been "
                f"given write access",
                None,
                ActionType.erasure,
                ExecutionLogStatus.error,
            )
            return 0

        formatted_input_data: NodeInput = self.pre_process_input_data(
            *inputs, group_dependent_fields=True
        )

        output = self.connector.mask_data(
            self.traversal_node,
            self.resources.policy,
            self.resources.request,
            retrieved_data,
            formatted_input_data,
        )
        self.log_end(ActionType.erasure)
        self.resources.cache_erasure(
            f"{self.key}", output
        )  # Cache that the erasure was performed in case we need to restart
        return output

    @retry(action_type=ActionType.consent, default_return=False)
    def consent_request(self, identity: Dict[str, Any]) -> bool:
        """Run consent request request"""

        if not self.can_write_data():
            logger.warning(
                "No consent on {} as its ConnectionConfig does not have write access.",
                self.traversal_node.node.address,
            )
            self.update_status(
                f"No values were erased since this connection {self.connector.configuration.key} has not been "
                f"given write access",
                None,
                ActionType.erasure,
                ExecutionLogStatus.error,
            )
            return False

        consent_preferences: List[Consent] = [
            Consent(**pref) for pref in self.resources.request.consent_preferences or []
        ]

        output: bool = self.connector.run_consent_request(
            self.traversal_node,
            self.resources.policy,
            self.resources.request,
            identity,
            consent_preferences,
        )
        self.log_end(ActionType.consent)
        return output


def collect_queries(
    traversal: Traversal, resources: TaskResources
) -> Dict[CollectionAddress, str]:
    """Collect all queries for dry-run"""

    def collect_queries_fn(
        tn: TraversalNode, data: Dict[CollectionAddress, str]
    ) -> None:
        if not tn.is_root_node():
            data[tn.address] = GraphTask(tn, resources).generate_dry_run_query()  # type: ignore

    env: Dict[CollectionAddress, str] = {}
    traversal.traverse(env, collect_queries_fn)
    return env


def update_mapping_from_cache(
    dsk: Dict[CollectionAddress, Tuple[Any, ...]],
    resources: TaskResources,
    start_fn: Callable,
) -> None:
    """When resuming a privacy request from a paused or failed state, update the `dsk` dictionary with results we've
    already obtained from a previous run. Remove upstream dependencies for these nodes, and just return the data we've
    already retrieved, rather than visiting them again.

    If there's no cached data, the dsk dictionary won't change.
    """

    cached_results: Dict[str, Optional[List[Row]]] = resources.get_all_cached_objects()

    for collection_name in cached_results:
        dsk[CollectionAddress.from_string(collection_name)] = (
            start_fn(cached_results[collection_name]),
        )


def start_function(seed: List[Dict[str, Any]]) -> Callable[[], List[Dict[str, Any]]]:
    """Return a function for collections with no upstream dependencies, that just start
    with seed data.

    This is used for root nodes or previously-visited nodes on restart."""

    def g() -> List[Dict[str, Any]]:
        return seed

    return g


async def run_access_request(
    privacy_request: PrivacyRequest,
    policy: Policy,
    graph: DatasetGraph,
    connection_configs: List[ConnectionConfig],
    identity: Dict[str, Any],
    session: Session,
) -> Dict[str, List[Row]]:
    """Run the access request"""
    traversal: Traversal = Traversal(graph, identity)
    with TaskResources(
        privacy_request, policy, connection_configs, session
    ) as resources:

        def collect_tasks_fn(
            tn: TraversalNode, data: Dict[CollectionAddress, GraphTask]
        ) -> None:
            """Run the traversal, as an action creating a GraphTask for each traversal_node."""
            if not tn.is_root_node():
                data[tn.address] = GraphTask(tn, resources)

        def termination_fn(
            *dependent_values: List[Row],
        ) -> Dict[str, Optional[List[Row]]]:
            """A termination function that just returns its inputs mapped to their source addresses.
            This needs to wait for all dependent keys because this is how dask is informed to wait for
            all terminating addresses before calling this."""

            return resources.get_all_cached_objects()

        env: Dict[CollectionAddress, Any] = {}
        end_nodes = traversal.traverse(env, collect_tasks_fn)

        dsk: Dict[CollectionAddress, Tuple[Any, ...]] = {
            k: (t.access_request, *t.input_keys) for k, t in env.items()
        }
        dsk[ROOT_COLLECTION_ADDRESS] = (start_function([traversal.seed_data]),)
        dsk[TERMINATOR_ADDRESS] = (termination_fn, *end_nodes)
        update_mapping_from_cache(dsk, resources, start_function)

        await fideslog_graph_rerun(
            prepare_rerun_graph_analytics_event(
                privacy_request, env, end_nodes, resources, ActionType.access
            )
        )
        privacy_request.cache_access_graph(format_graph_for_caching(env, end_nodes))

        v = delayed(get(dsk, TERMINATOR_ADDRESS, num_workers=1))
        return v.compute()


def get_cached_data_for_erasures(
    privacy_request_id: str,
) -> Dict[str, Any]:
    """
    Fetches processed access request results to be used for erasures.

    Processing may have added indicators to not mask certain elements in array data.
    """
    cache = get_cache()
    value_dict = cache.get_encoded_objects_by_prefix(
        f"PLACEHOLDER_RESULTS__{privacy_request_id}"
    )
    return {k.split("__")[-1]: v for k, v in value_dict.items()}


def update_erasure_mapping_from_cache(
    dsk: Dict[CollectionAddress, Tuple[Any, ...]],
    resources: TaskResources,
    start_fn: Callable,
) -> None:
    """On pause or restart from failure, update the dsk graph to skip running erasures on collections
    we've already visited. Instead, just return the previous count of rows affected.

    If there's no cached data, the dsk dictionary won't change.
    """
    cached_erasures: Dict[str, int] = resources.get_all_cached_erasures()

    for collection_name in cached_erasures:
        dsk[CollectionAddress.from_string(collection_name)] = (
            start_fn(cached_erasures[collection_name]),
        )


async def run_erasure(  # pylint: disable = too-many-arguments
    privacy_request: PrivacyRequest,
    policy: Policy,
    graph: DatasetGraph,
    connection_configs: List[ConnectionConfig],
    identity: Dict[str, Any],
    access_request_data: Dict[str, List[Row]],
    session: Session,
) -> Dict[str, int]:
    """Run an erasure request"""
    traversal: Traversal = Traversal(graph, identity)
    with TaskResources(
        privacy_request, policy, connection_configs, session
    ) as resources:

        def collect_tasks_fn(
            tn: TraversalNode, data: Dict[CollectionAddress, GraphTask]
        ) -> None:
            """Run the traversal, as an action creating a GraphTask for each traversal_node."""
            if not tn.is_root_node():
                data[tn.address] = GraphTask(tn, resources)

        env: Dict[CollectionAddress, Any] = {}
        end_nodes = traversal.traverse(env, collect_tasks_fn)

        def termination_fn(*dependent_values: int) -> Tuple[int, ...]:
            """The dependent_values here is an int output from each task feeding in, where
            each task reports the output of 'task.rtf(access_request_data)', which is the number of
            records updated.

            The termination function just returns this tuple of ints."""
            return dependent_values

        access_request_data[ROOT_COLLECTION_ADDRESS.value] = [identity]

        dsk: Dict[CollectionAddress, Any] = {
            k: (
                t.erasure_request,
                access_request_data.get(
                    str(k), []
                ),  # Pass in the results of the access request for this collection
                *[
                    access_request_data.get(
                        str(upstream_key), []
                    )  # Additionally pass in the original input data we used for the access request. It's helpful in
                    # cases like the EmailConnector where the access request doesn't actually retrieve data.
                    for upstream_key in t.input_keys
                ],
            )
            for k, t in env.items()
        }
        # terminator function waits for all keys
        dsk[TERMINATOR_ADDRESS] = (termination_fn, *env.keys())
        update_erasure_mapping_from_cache(dsk, resources, start_function)
        await fideslog_graph_rerun(
            prepare_rerun_graph_analytics_event(
                privacy_request, env, end_nodes, resources, ActionType.erasure
            )
        )
        v = delayed(get(dsk, TERMINATOR_ADDRESS, num_workers=1))

        update_cts: Tuple[int, ...] = v.compute()
        # we combine the output of the termination function with the input keys to provide
        # a map of {collection_name: records_updated}:
        erasure_update_map: Dict[str, int] = dict(
            zip([str(x) for x in env], update_cts)
        )

        return erasure_update_map


async def run_consent_request(  # pylint: disable = too-many-arguments
    privacy_request: PrivacyRequest,
    policy: Policy,
    graph: DatasetGraph,
    connection_configs: List[ConnectionConfig],
    identity: Dict[str, Any],
    session: Session,
) -> Dict[str, bool]:
    """Run a consent request

    The graph built is very simple: there are no relationships between the nodes, every node has
    identity data input and every node outputs whether the consent request succeeded.

    The DatasetGraph passed in is expected to have one Node per Dataset.  That Node is expected to carry out requests
    for the Dataset as a whole.
    """

    with TaskResources(
        privacy_request, policy, connection_configs, session
    ) as resources:
        graph_keys: List[CollectionAddress] = list(graph.nodes.keys())
        dsk: Dict[CollectionAddress, Any] = {}

        for col_address, node in graph.nodes.items():
            traversal_node = TraversalNode(node)
            task = GraphTask(traversal_node, resources)
            dsk[col_address] = (task.consent_request, identity)

        def termination_fn(*dependent_values: bool) -> Tuple[bool, ...]:
            """The dependent_values here is an bool output from each task feeding in, where
            each task reports the output of 'task.consent_request(identity_data)', which is whether the
            consent request succeeded

            The termination function just returns this tuple of booleans."""
            return dependent_values

        # terminator function waits for all keys
        dsk[TERMINATOR_ADDRESS] = (termination_fn, *graph_keys)

        v = delayed(get(dsk, TERMINATOR_ADDRESS, num_workers=1))

        update_successes: Tuple[bool, ...] = v.compute()
        # we combine the output of the termination function with the input keys to provide
        # a map of {collection_name: whether consent request succeeded}:
        consent_update_map: Dict[str, bool] = dict(
            zip([coll.value for coll in graph_keys], update_successes)
        )

        return consent_update_map


def build_affected_field_logs(
    node: Node, policy: Policy, action_type: ActionType
) -> List[Dict[str, Any]]:
    """For a given node (collection), policy, and action_type (access or erasure) format all of the fields that
    were potentially touched to be stored in the ExecutionLogs for troubleshooting.

    :Example:
    [{
        "path": "dataset_name:collection_name:field_name",
        "field_name": "field_name",
        "data_categories": ["data_category_1", "data_category_2"]
    }]
    """

    targeted_field_paths: Dict[FieldAddress, str] = {}

    for rule in policy.rules:  # type: ignore[attr-defined]
        if rule.action_type != action_type:
            continue
        rule_categories: List[str] = rule.get_target_data_categories()
        if not rule_categories:
            continue

        collection_categories: Dict[
            str, List[FieldPath]
        ] = node.collection.field_paths_by_category  # type: ignore
        for rule_cat in rule_categories:
            for collection_cat, field_paths in collection_categories.items():
                if collection_cat.startswith(rule_cat):
                    targeted_field_paths.update(
                        {
                            node.address.field_address(field_path): collection_cat
                            for field_path in field_paths
                        }
                    )

    ret: List[Dict[str, Any]] = []
    for field_address, data_categories in targeted_field_paths.items():
        ret.append(
            {
                "path": field_address.value,
                "field_name": field_address.field_path.string_path,
                "data_categories": [data_categories],
            }
        )

    return ret



from fides.api.ops.schemas.base_class import BaseSchema


class IdentityVerificationConfigResponse(BaseSchema):
    """Response for identity verification config info"""

    identity_verification_required: bool
    valid_email_config_exists: bool



"""add audit log

Revision ID: c98da12d76f8
Revises: 29a7d707163a
Create Date: 2022-05-04 20:16:39.246537

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "c98da12d76f8"
down_revision = "29a7d707163a"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "auditlog",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("user_id", sa.String(), nullable=True),
        sa.Column("privacy_request_id", sa.String(), nullable=True),
        sa.Column(
            "action",
            sa.Enum("approved", "denied", name="auditlogaction"),
            nullable=False,
        ),
        sa.Column("message", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_auditlog_action"), "auditlog", ["action"], unique=False)
    op.create_index(op.f("ix_auditlog_id"), "auditlog", ["id"], unique=False)
    op.create_index(
        op.f("ix_auditlog_privacy_request_id"),
        "auditlog",
        ["privacy_request_id"],
        unique=False,
    )
    op.create_index(op.f("ix_auditlog_user_id"), "auditlog", ["user_id"], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_auditlog_user_id"), table_name="auditlog")
    op.drop_index(op.f("ix_auditlog_privacy_request_id"), table_name="auditlog")
    op.drop_index(op.f("ix_auditlog_id"), table_name="auditlog")
    op.drop_index(op.f("ix_auditlog_action"), table_name="auditlog")
    op.drop_table("auditlog")
    # ### end Alembic commands ###



from typing import Any, Dict, List, Optional

from loguru import logger
from sqlalchemy.orm import Session

from fides.api.ops.common_exceptions import (
    MessageDispatchException,
    PrivacyRequestErasureEmailSendRequired,
)
from fides.api.ops.graph.config import CollectionAddress, FieldPath
from fides.api.ops.graph.traversal import TraversalNode
from fides.api.ops.models.connectionconfig import (
    ConnectionConfig,
    ConnectionTestStatus,
    ConnectionType,
)
from fides.api.ops.models.datasetconfig import DatasetConfig
from fides.api.ops.models.messaging import MessagingConfig
from fides.api.ops.models.policy import CurrentStep, Policy, Rule
from fides.api.ops.models.privacy_request import (
    CheckpointActionRequired,
    ManualAction,
    PrivacyRequest,
)
from fides.api.ops.schemas.connection_configuration import EmailSchema
from fides.api.ops.schemas.messaging.messaging import (
    MessagingActionType,
    MessagingServiceType,
)
from fides.api.ops.schemas.redis_cache import Identity
from fides.api.ops.service.connectors.base_connector import BaseConnector
from fides.api.ops.service.connectors.query_config import ManualQueryConfig
from fides.api.ops.service.messaging.message_dispatch_service import dispatch_message
from fides.api.ops.util.collection_util import Row, append
from fides.lib.models.audit_log import AuditLog, AuditLogAction


class EmailConnector(BaseConnector[None]):
    def query_config(self, node: TraversalNode) -> ManualQueryConfig:
        """Intentionally reusing the ManualQueryConfig here. We're only using methods off of the base
        QueryConfig class here.
        """
        return ManualQueryConfig(node)

    def create_client(self) -> None:
        """Stub"""

    def close(self) -> None:
        """N/A for the EmailConnector"""

    def test_connection(self) -> Optional[ConnectionTestStatus]:
        """
        Sends an email to the "test_email" configured, just to establish that the email workflow is working.
        """
        config = EmailSchema(**self.configuration.secrets or {})
        logger.info("Starting test connection to {}", self.configuration.key)

        db = Session.object_session(self.configuration)

        email_service: Optional[str] = _get_email_messaging_config_service_type(db=db)

        try:
            # synchronous for now since failure to send is considered a connection test failure
            dispatch_message(
                db=db,
                action_type=MessagingActionType.MESSAGE_ERASURE_REQUEST_FULFILLMENT,
                to_identity=Identity(email=config.test_email),
                service_type=email_service,
                message_body_params=[
                    CheckpointActionRequired(
                        step=CurrentStep.erasure,
                        collection=CollectionAddress("test_dataset", "test_collection"),
                        action_needed=[
                            ManualAction(
                                locators={"id": ["example_id"]},
                                get=None,
                                update={
                                    "test_field": "null_rewrite",
                                },
                            )
                        ],
                    )
                ],
            )
        except MessageDispatchException as exc:
            logger.info("Email connector test failed with exception {}", exc)
            return ConnectionTestStatus.failed
        return ConnectionTestStatus.succeeded

    def retrieve_data(  # type: ignore
        self,
        node: TraversalNode,
        policy: Policy,
        privacy_request: PrivacyRequest,
        input_data: Dict[str, List[Any]],
    ) -> Optional[List[Row]]:
        """Access requests are not supported at this time."""
        logger.info(
            "Access requests not supported for email connector '{}' at this time.",
            node.address.value,
        )
        return []

    def mask_data(  # type: ignore
        self,
        node: TraversalNode,
        policy: Policy,
        privacy_request: PrivacyRequest,
        rows: List[Row],
        input_data: Dict[str, List[Any]],
    ) -> None:
        """Cache instructions for how to mask data in this collection.
        One email will be sent for all collections in this dataset at the end of the privacy request execution.
        """

        manual_action: ManualAction = self.build_masking_instructions(
            node, policy, input_data
        )

        logger.info("Caching action needed for collection: '{}", node.address.value)
        privacy_request.cache_email_connector_template_contents(
            step=CurrentStep.erasure,
            collection=node.address,
            action_needed=[manual_action],
        )

        # Raises a special exception just to update the ExecutionLog message.  The email send itself
        # is postponed until all collections have been visited.
        raise PrivacyRequestErasureEmailSendRequired("email prepared")

    def build_masking_instructions(
        self, node: TraversalNode, policy: Policy, input_data: Dict[str, List[Any]]
    ) -> ManualAction:
        """
        Generate information on how to find and mask relevant records on this collection.

        Because we don't have any rows from a completed "access request" for this source, we pass on both "access"
        and "erasure" instructions to the source.
        """

        # Values for locating relevant records on the current collection.
        locators: Dict[str, Any] = node.typed_filtered_values(input_data)
        # Add additional locators from the same dataset for which we don't have data
        for edge in node.incoming_edges_from_same_dataset():
            append(locators, edge.f2.field_path.string_path, str(edge.f1))

        # Build which fields to mask and how to mask
        fields_to_mask_by_rule: Dict[Rule, List[FieldPath]] = self.query_config(
            node
        ).build_rule_target_field_paths(policy)
        mask_map: Dict[str, Any] = {}
        for rule, field_paths in fields_to_mask_by_rule.items():
            for rule_field_path in field_paths:
                # Map field paths to the specified masking strategy.  If multiple rules target
                # the same field path, the last one will win
                mask_map[rule_field_path.string_path] = (
                    rule.masking_strategy.get("strategy")
                    if rule.masking_strategy
                    else None
                )

        # Returns a ManualAction even if there are no fields to mask on this collection,
        # because the locators still may be needed to find data to mask on dependent collections
        return ManualAction(locators=locators, update=mask_map if mask_map else None)


def email_connector_erasure_send(db: Session, privacy_request: PrivacyRequest) -> None:
    """
    Send emails to configured third-parties with instructions on how to erase remaining data.
    Combined all the collections on each email-based dataset into one email.
    """
    email_dataset_configs = db.query(DatasetConfig, ConnectionConfig).filter(
        DatasetConfig.connection_config_id == ConnectionConfig.id,
        ConnectionConfig.connection_type == ConnectionType.email,
    )
    for ds, cc in email_dataset_configs:
        template_values: List[
            CheckpointActionRequired
        ] = privacy_request.get_email_connector_template_contents_by_dataset(
            CurrentStep.erasure, ds.ctl_dataset.fides_key
        )

        if not template_values:
            logger.info(
                "No email sent: no template values saved for '{}'",
                ds.ctl_dataset.fides_key,
            )
            return

        if not any(
            (
                action_required.action_needed[0].update
                if action_required and action_required.action_needed
                else False
                for action_required in template_values
            )
        ):
            logger.info(
                "No email sent: no masking needed on '{}'", ds.ctl_dataset.fides_key
            )
            return

        email_service: Optional[str] = _get_email_messaging_config_service_type(db=db)

        dispatch_message(
            db,
            action_type=MessagingActionType.MESSAGE_ERASURE_REQUEST_FULFILLMENT,
            to_identity=Identity(email=cc.secrets.get("to_email")),
            service_type=email_service,
            message_body_params=template_values,
        )

        logger.info(
            "Email send succeeded for request '{}' for dataset: '{}'",
            privacy_request.id,
            ds.ctl_dataset.fides_key,
        )
        AuditLog.create(
            db=db,
            data={
                "user_id": "system",
                "privacy_request_id": privacy_request.id,
                "action": AuditLogAction.email_sent,
                "message": f"Erasure email instructions dispatched for '{ds.ctl_dataset.fides_key}'",
            },
        )


def _get_email_messaging_config_service_type(db: Session) -> Optional[str]:
    """
    Email connectors require that an email messaging service has been configured.
    Prefers Twilio if both Twilio email AND Mailgun has been configured.
    """
    messaging_configs: Optional[List[MessagingConfig]] = MessagingConfig.query(
        db=db
    ).all()
    if not messaging_configs:
        # let messaging dispatch service handle non-existent service
        return None
    twilio_email_config = next(
        (
            config
            for config in messaging_configs
            if config.service_type == MessagingServiceType.TWILIO_EMAIL
        ),
        None,
    )
    mailgun_config = next(
        (
            config
            for config in messaging_configs
            if config.service_type == MessagingServiceType.MAILGUN
        ),
        None,
    )
    if twilio_email_config:
        # we prefer twilio over mailgun
        return MessagingServiceType.TWILIO_EMAIL.value
    if mailgun_config:
        return MessagingServiceType.MAILGUN.value
    return None



from enum import Enum
from typing import List, Optional

from pydantic import BaseModel

from fides.api.ops.models.policy import WebhookDirection
from fides.api.ops.schemas.redis_cache import Identity


class CallbackType(Enum):
    """We currently have two types of Policy Webhooks: pre and post"""

    pre = "pre"
    post = "post"


class SecondPartyRequestFormat(BaseModel):
    """The request body we will use when calling a user's HTTP endpoint from fides.api"""

    privacy_request_id: str
    direction: WebhookDirection
    callback_type: CallbackType
    identity: Identity

    class Config:
        """Using enum values"""

        use_enum_values = True


class SecondPartyResponseFormat(BaseModel):
    """The expected response from a user's HTTP endpoint that receives callbacks from fides.api

    Responses are only expected (and considered) for two_way webhooks.
    """

    derived_identity: Optional[Identity] = None
    halt: bool

    class Config:
        """Using enum values"""

        use_enum_values = True


class PrivacyRequestResumeFormat(BaseModel):
    """Expected request body to resume a privacy request after it was paused by a webhook"""

    derived_identity: Optional[Identity] = {}  # type: ignore

    class Config:
        """Using enum values"""

        use_enum_values = True


class WebhookJWE(BaseModel):
    """Describes JWE that is given to the user that they need to send with their request
    to resume a privacy request"""

    webhook_id: str
    scopes: List[str]
    iat: str



from __future__ import annotations

import abc
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Extra, root_validator

from fides.api.ops.models.connectionconfig import ConnectionTestStatus
from fides.api.ops.schemas import Msg


class ConnectionConfigSecretsSchema(BaseModel, abc.ABC):
    """Abstract Base Schema for updating Connection Configuration Secrets"""

    url: Optional[str] = None  # User can always specify the URL directly

    _required_components: List[str]

    def __init_subclass__(cls: BaseModel, **kwargs: Any):  # type: ignore
        super().__init_subclass__(**kwargs)  # type: ignore
        if not getattr(cls, "_required_components"):
            raise TypeError(f"Class {cls.__name__} must define '_required_components.'")  # type: ignore

    @root_validator
    @classmethod
    def required_components_supplied(  # type: ignore
        cls: ConnectionConfigSecretsSchema, values: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Validate that the minimum required components have been supplied.

        Connection configurations either 1) need the entire URL
        *OR* 2) all of the required components to *build* that URL."""
        min_fields_present = values.get("url") or all(
            values.get(component) for component in cls._required_components
        )
        if not min_fields_present:
            raise ValueError(
                f"{cls.__name__} must be supplied a 'url' or all of: {cls._required_components}."  # type: ignore
            )

        return values

    class Config:
        """Only permit selected secret fields to be stored."""

        extra = Extra.forbid
        orm_mode = True


class TestStatusMessage(Msg):
    """A schema for checking status."""

    test_status: Optional[ConnectionTestStatus] = None
    failure_reason: Optional[str] = None



"""add authorization request

Revision ID: 1ff88b7bd579
Revises: 3a7c5fb119c9
Create Date: 2022-05-25 04:09:22.149110

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "1ff88b7bd579"
down_revision = "3a7c5fb119c9"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "authenticationrequest",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("connection_key", sa.String(), nullable=False),
        sa.Column("state", sa.String(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("connection_key"),
    )
    op.create_index(
        op.f("ix_authenticationrequest_id"),
        "authenticationrequest",
        ["id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_authenticationrequest_state"),
        "authenticationrequest",
        ["state"],
        unique=True,
    )


def downgrade():
    op.drop_index(
        op.f("ix_authenticationrequest_state"), table_name="authenticationrequest"
    )
    op.drop_index(
        op.f("ix_authenticationrequest_id"), table_name="authenticationrequest"
    )
    op.drop_table("authenticationrequest")



from __future__ import annotations

from typing import Any

from fides.api.ops.common_exceptions import NoSuchConnectionTypeSecretSchemaError
from fides.api.ops.models.connectionconfig import ConnectionType
from fides.api.ops.schemas.connection_configuration import (
    SaaSSchemaFactory,
    secrets_schemas,
)
from fides.api.ops.schemas.connection_configuration.connection_config import (
    ConnectionSystemTypeMap,
    SystemType,
)
from fides.api.ops.schemas.saas.saas_config import SaaSConfig
from fides.api.ops.service.connectors.saas.connector_registry_service import (
    load_registry,
    registry_file,
)
from fides.api.ops.util.saas_util import encode_file_contents, load_config


def connection_type_secret_schema(*, connection_type: str) -> dict[str, Any]:
    """Returns the secret fields that should be supplied to authenticate with a particular connection type.

    Note that this does not return actual secrets, instead we return the *types* of
    secret fields needed to authenticate.
    """
    connection_system_types: list[ConnectionSystemTypeMap] = get_connection_types()
    if not any(item.identifier == connection_type for item in connection_system_types):
        raise NoSuchConnectionTypeSecretSchemaError(
            f"No connection type found with name '{connection_type}'."
        )

    if connection_type in [db_type.value for db_type in ConnectionType]:
        return secrets_schemas[connection_type].schema()

    config = SaaSConfig(**load_config(f"data/saas/config/{connection_type}_config.yml"))
    schema = SaaSSchemaFactory(config).get_saas_schema().schema()

    # rearrange the default order of the properties generated by Pydantic
    # to reflect the order defined in the connector_params and external_references
    order = [connector_param.name for connector_param in config.connector_params]
    if config.external_references:
        order.extend(
            [
                external_reference.name
                for external_reference in config.external_references
            ]
        )

    schema["properties"] = {
        name: value
        for name, value in sorted(
            schema["properties"].items(), key=lambda item: order.index(item[0])
        )
    }
    return schema


def get_connection_types(
    search: str | None = None, system_type: SystemType | None = None
) -> list[ConnectionSystemTypeMap]:
    def is_match(elem: str) -> bool:
        """If a search query param was included, is it a substring of an available connector type?"""
        return search.lower() in elem.lower() if search else True

    connection_system_types: list[ConnectionSystemTypeMap] = []
    if system_type == SystemType.database or system_type is None:
        database_types: list[str] = sorted(
            [
                conn_type.value
                for conn_type in ConnectionType
                if conn_type
                not in [
                    ConnectionType.saas,
                    ConnectionType.https,
                    ConnectionType.manual,
                    ConnectionType.email,
                    ConnectionType.manual_webhook,
                    ConnectionType.fides,
                ]
                and is_match(conn_type.value)
            ]
        )
        connection_system_types.extend(
            [
                ConnectionSystemTypeMap(
                    identifier=item,
                    type=SystemType.database,
                    human_readable=ConnectionType(item).human_readable,
                )
                for item in database_types
            ]
        )
    if system_type == SystemType.saas or system_type is None:
        registry = load_registry(registry_file)
        saas_types: list[str] = sorted(
            [
                saas_type
                for saas_type in registry.connector_types()
                if is_match(saas_type)
            ]
        )

        for item in saas_types:
            if registry.get_connector_template(item) is not None:
                connector_template = registry.get_connector_template(  # type: ignore[union-attr]
                    item
                )

            connection_system_types.append(
                ConnectionSystemTypeMap(
                    identifier=item,
                    type=SystemType.saas,
                    human_readable=connector_template.human_readable
                    if connector_template is not None
                    else "",
                    encoded_icon=encode_file_contents(connector_template.icon)
                    if connector_template is not None
                    else None,
                )
            )

    if system_type == SystemType.manual or system_type is None:
        manual_types: list[str] = sorted(
            [
                manual_type.value
                for manual_type in ConnectionType
                if manual_type == ConnectionType.manual_webhook
                and is_match(manual_type.value)
            ]
        )
        connection_system_types.extend(
            [
                ConnectionSystemTypeMap(
                    identifier=item,
                    type=SystemType.manual,
                    human_readable=ConnectionType(item).human_readable,
                )
                for item in manual_types
            ]
        )
    return connection_system_types



"""
Contains api endpoints for fides web pages
"""
from fastapi import Depends, Security
from fastapi.responses import HTMLResponse
from sqlalchemy.ext.asyncio import AsyncSession

from fides.api.ctl.database.session import get_async_db
from fides.api.ctl.routes.crud import list_resource  # type: ignore[attr-defined]
from fides.api.ctl.sql_models import Evaluation  # type: ignore[attr-defined]
from fides.api.ctl.utils.api_router import APIRouter
from fides.api.ops.api.v1 import scope_registry
from fides.api.ops.util.oauth_util import verify_oauth_client_cli

router = APIRouter(
    tags=["View"],
    prefix="/view",
)


@router.get(
    "/evaluations",
    dependencies=[
        Security(verify_oauth_client_cli, scopes=[scope_registry.CLI_OBJECTS_READ])
    ],
)
async def evaluation_view(db: AsyncSession = Depends(get_async_db)) -> HTMLResponse:
    "Returns an html document with a list of evaluations"

    html = '<html lang="en">'
    html += "<head>"
    html += '<link rel="stylesheet" href="https://stackpath.bootstrapcdn.com/bootstrap/4.3.1/css/bootstrap.min.css">'
    html += "<body>"
    html += "<h2>Evaluations</h2>"
    html += '<table class="table table-striped" style="text-align: left;">'
    html += "<tr><th>Id</th><th>Status</th><th>Violations</th><th>Message</th></tr>"
    for evaluation in await list_resource(Evaluation, db):
        html += "<tr>"
        html += f"<td>{evaluation.fides_key}</td>"
        html += f"<td>{evaluation.status}</td>"

        html += "<td>"
        for index, violation in enumerate(evaluation.violations):
            violating_attributes = violation.get("violating_attributes")
            html += f"<b>Violation {index + 1}</b><br/>"
            html += "<ul>"
            html += f"<li>Data Categories: {violating_attributes.get('data_categories')}</li>"
            html += (
                f"<li>Data Subjects: {violating_attributes.get('data_subjects')}</li>"
            )
            html += f"<li>Data Uses: {violating_attributes.get('data_uses')}</li>"
            html += (
                f"<li>Data Qualifier: {violating_attributes.get('data_qualifier')}</li>"
            )
            html += f"<li>Detail: {violation.get('detail')}</li>"
            html += "</ul>"
        html += "</td>"

        html += f"<td>{evaluation.message}</td>"
        html += "</tr>"
    html += "</table>"
    html += "</head>"
    html += "</html>"
    return HTMLResponse(html)
